# -*- coding: utf-8 -*-

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.px import Px
from appy import utils
from appy.model.fields import Field
from appy.ui.layout import Layouts, Layout
from appy.database.operators import Operator

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
emptyList = []

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
RADIOS_KO  = 'You cannot render this field as radio buttons if it may ' \
             'contain several values (max multiplicity is higher than 1).'
CBS_KO     = 'You cannot render this field as checkboxes if it can ' \
             'only contain a single value (max multiplicity is 1).'
IEDIT_KO   = 'It is currently not possible to inline-edit a multi-' \
             'valued Select field.'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Selection:
    '''If you want to have dynamically computed possible values for a Select
       field, use a Selection instance.'''

    def __init__(self, method, single=None):
        # p_method must be a method that will be called every time Appy will
        # need to get the list of possible values for the related field.
        #  - p_method can be the method object in itself, or the method name.
        #  - it can be (or correspond to) an instance method of the class
        #    defining the related field; it can also refer to a method defined
        #    on the tool. In this latter case, you are forced to define a method
        #    name (not a method object), prefixed with "tool:".

        # m_method, in most cases, accepts no arg and return a list (or tuple)
        # of pairs (lists or tuples): (id, text), where:
        #   - "id" is one of the possible values for the field;
        #   - "text" is the value as will be shown in the UI.
        # You can nevertheless specify args to the method. In that case, specify
        # the method name, followed by args separated with stars, like in
        #
        #                      method1*arg1*arg2
        #
        # Only string args are supported.
        # Within m_method, you will usually call the standard translation method
        # (m_translate) to produce an i18n version of "text".
        self.method = method

        # While p_method is called on edit and filter layouts to list the
        # possible values to select, the method you specify in p_single allows
        # to produce the translated text for a single value. It accepts a single
        # arg, being one of the values as stored by the select field, and must
        # return its translated text. If you don't specify a p_single method,
        # p_method will be used instead, but useless processing will take place,
        # because p_method produces texts for the complete list of possible
        # values. Moreover, a stored select value may not be among the list of
        # values as returned by p_method anymore.
        self.single = single

    def getText(self, o, value, field, language=None):
        '''Gets the text that corresponds to p_value'''
        # Use method p_single if specified
        if self.single: return self.single(o, value)
        # Use p_self.method
        withTranslations = language if language else True
        vals = field.getPossibleValues(o, ignoreMasterValues=True,
                                       withTranslations=withTranslations)
        for v, text in vals:
            if v == value: return text
        return value

    def getValues(self, o, className=None):
        '''Gets the translated values in this selection'''
        method = self.method
        # Call self.method for getting the (dynamic) values
        if isinstance(method, str):
            # This is the name of a method, not the method itself. Unwrap args
            # if any.
            if method.find('*') != -1:
                elems = method.split('*')
                method = elems[0]
                args = elems[1:]
            else:
                args = []
            # On what object must we call the method ?
            if method.startswith('tool:'):
                o = o.tool
                method = method[5:]
            # Get the method from its name
            method = getattr(o, method)
        else:
            args = [o]
        # If a p_className is given, transmit the corresponding class as method
        # arg.
        if className:
            args.append(o.model.classes.get(className))
        # Call the method
        try:
            r = method(*args)
        except TypeError:
            args.pop()
            r = method(*args)
        return r

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Select(Field):
    '''Field allowing to choose a value among a list of possible values. Each
       value is represented and stored as a string.'''

    class Layouts(Layouts):
        '''Select-specific layouts'''
        d  = Layouts(Layout('d2-l-f;rv=', width=None))
        g  = Layouts(edit=Layout('f;rv=', width=None),
                     view=Layout('fl', width=None))
        gd = Layouts(edit=Layout('d2-f;rv=', width=None),
                     view=Layout('fl', width=None))

        @classmethod
        def getDefault(class_, field):
            '''Default layouts for this Select p_field'''
            return class_.g if field.inGrid() else class_.b

    view = Px('''
     <!-- No value at all -->
     <span if="not value" class="smaller">-</span>
     <!-- A single value -->
     <x if="value and not isMultiple">::field.getInlineEditableValue(o, \
                                          value, layout, name=name)</x>
     <!-- Several values -->
     <ul if="value and isMultiple"><li for="sv in value"><i>::sv</i></li></ul>
     <!-- If this field is a master field -->
     <input type="hidden" if="masterCss" class=":masterCss" value=":rawValue"
            name=":name" id=":name"/>''')

    # More compact representation on the cell layout
    cell = Px('''
     <x var="multiple=value and isMultiple">
      <x if="multiple">:', '.join(value)</x>
      <x if="not multiple">:field.view</x>
     </x>''')

    edit = Px('''
     <x var="isSelect=field.render == 'select';
             possibleValues=field.getPossibleValues(o, withTranslations=True, \
               withBlankValue=isSelect);
             charsWidth=field.getWidthInChars(False)">
     <select if="isSelect" name=":name" id=":name" class=":masterCss"
       multiple=":isMultiple" onchange=":field.getOnChange(o, layout)"
       size=":field.getSelectSize(False, isMultiple)"
       style=":field.getSelectStyle(False, isMultiple)">
      <option for="val, text in possibleValues" value=":val"
              selected=":field.isSelected(o, name, val, rawValue)"
              title=":text">:Px.truncateValue(text, charsWidth)</option>
     </select>
     <x if="not isSelect">
      <div for="val, text in possibleValues"
           var2="vid='%s_%s' % (name, val)">
       <input type=":field.render" name=":name" id=":vid" value=":val"
              class=":masterCss" onchange=":field.getOnChange(o, layout)"
              checked=":field.isSelected(o, name, val, rawValue)"/>
       <label lfor=":vid" class="subLabel">:text</label>
      </div>
     </x>
     <script if="hostLayout" var2="x=o.Lock.set(o, user, field=field)">:\
      'prepareForAjaxSave(%s,%s,%s,%s)' % \
       (q(name), q(o.iid), q(o.url), q(hostLayout))</script></x>''')

    # On the search form, show a multi-selection widget with a "AND/OR" selector
    search = Px('''
     <!-- The "and" / "or" radio buttons -->
     <x if="field.multiplicity[1] != 1"
        var2="operName='o_%s' % name;
              orName='%s_or' % operName;
              andName='%s_and' % operName">
      <input type="radio" name=":operName" id=":orName"
             checked="checked" value="or"/>
      <label lfor=":orName">:_('search_or')</label>
      <input type="radio" name=":operName" id=":andName" value="and"/>
      <label lfor=":andName">:_('search_and')</label><br/>
     </x>

     <!-- The list of values -->
     <select var="preSelected=field.sdefault;
                  charsWidth=field.getWidthInChars(True)"
       name=":widgetName" multiple="multiple"
       size=":field.getSelectSize(True, True)"
       style=":field.getSelectStyle(True, True)"
       onchange=":field.getOnChange(tool, 'search', className)">
      <option for="val, text in field.getPossibleValues(tool, \
            withTranslations=True, withBlankValue=False, className=className)"
        selected=":val in preSelected" value=":val"
        title=":text">:Px.truncateValue(text, charsWidth)</option>
     </select><br/>''')

    # Widget for filtering object values on search results
    pxFilter = Px('''
     <select var="name=field.name;
                  filterId='%s_%s' % (mode.hook, name);
                  charsWidth=field.getWidthInChars(True)"
        id=":filterId" name=":filterId" class="discreet"
          onchange=":'askBunchFiltered(%s,%s)' % (q(mode.hook), q(name))">
      <option for="val, text in field.getPossibleValues(tool, \
                withTranslations=True, withBlankValue='forced', \
                blankLabel='everything', className=className)"
              selected=":mode.inFilter(name, val)" value=":val"
              title=":text">:Px.truncateValue(text, charsWidth)</option>
     </select>''')

    def __init__(self, validator=None, multiplicity=(0,1), default=None,
      defaultOnEdit=None, show=True, page='main', group=None, layouts=None,
      move=0, indexed=False, mustIndex=True, indexValue=None,
      emptyIndexValue='-', searchable=False, filterField=None,
      readPermission='read', writePermission='write', width=None, height=None,
      maxChars=None, colspan=1, master=None, masterValue=None, focus=False,
      historized=False, mapping=None, generateLabel=None, label=None,
      sdefault='', scolspan=1, swidth=None, sheight=None, persist=True,
      inlineEdit=False, view=None, cell=None, edit=None, xml=None,
      translations=None, noValueLabel='choose_a_value', render='select',
      svalidator=None):
        # When choosing a value in a select widget, the entry representing no
        # value is translated according the label defined in attribute
        # "noValueLabel". The default one is something like "[ choose ]", but if
        # you prefer a less verbose version, you can use "no_value" that simply
        # displays a dash, or your own label.
        self.noValueLabel = noValueLabel
        # A Select field, is, by default, rendered as a HTML select widget, but
        # there are alternate render modes. Here are all the possible render
        # modes.
        # ----------------------------------------------------------------------
        #  render    | The field is rendered as...
        # ----------------------------------------------------------------------
        # "select"   | an HTML select widget, rendered as a dropdown list if
        #            | max multiplicity is 1 or as a selection box if several
        #            | values can be chosen.
        # ----------------------------------------------------------------------
        # "radio"    | radio buttons. This mode is valid for fields with a max
        #            | multiplicity being 1.
        # ----------------------------------------------------------------------
        # "checkbox" | checkboxes. This mode is valid for fields with a max
        #            | multiplicity being higher than 1.
        # ----------------------------------------------------------------------
        self.render = render
        # Call the base constructor
        Field.__init__(self, validator, multiplicity, default, defaultOnEdit,
          show, page, group, layouts, move, indexed, mustIndex, indexValue,
          emptyIndexValue, searchable, filterField, readPermission,
          writePermission, width, height, maxChars, colspan, master,
          masterValue, focus, historized, mapping, generateLabel, label,
          sdefault, scolspan, swidth, sheight, persist, inlineEdit, view, cell,
          edit, xml, translations)
        # Default value on the *s*earch form. It must be a list of values.
        self.sdefault = sdefault or []
        # Validator to use on the *s*earch form. This has sense only if you use
        # a Selection instance as validator.
        self.svalidator = svalidator or validator
        # Default width, height and maxChars
        if width is None:
            self.width = 30
        if height is None:
            self.height = 4
        # Define the filter PX when appropriate
        if self.indexed:
            self.filterPx = 'pxFilter'
        self.swidth = self.swidth or self.width
        self.sheight = self.sheight or self.height
        self.checkParameters()

    def checkParameters(self):
        '''Ensure coherence between parameters'''
        # Valid render modes depend on multiplicities
        multiple = self.isMultiValued()
        if multiple and (self.render == 'radio'):
            raise Exception(RADIOS_KO)
        if not multiple and (self.render == 'checkbox'):
            raise Exception(CBS_KO)
        # It is currently not possible to inline-edit a multi-valued field
        if multiple and self.inlineEdit:
            raise Exception(IEDIT_KO)

    def isSelected(self, o, name, possibleValue, dbValue):
        '''When displaying a Select field, must the p_possibleValue appear as
           selected? p_name is given and used instead of field.name because it
           may contain a row number from a field within a List field.'''
        req = o.req
        # Get the value we must compare (from request or from database)
        if name in req:
            compValue = req[name]
        else:
            compValue = dbValue
        # Compare the value
        if type(compValue) in utils.sequenceTypes:
            return possibleValue in compValue
        return possibleValue == compValue

    def getValue(self, o, name=None, layout=None, single=None):
        value = Field.getValue(self, o, name, layout, single)
        if not value:
            return emptyList if self.isMultiValued() else value
        if isinstance(value, str) and self.isMultiValued():
            value = [value]
        elif isinstance(value, tuple):
            value = list(value)
        return value

    def getTranslatedValue(self, o, value, language=None):
        '''Get the translated text for p_value, when p_value is one of p_self's
           authorized "static" values according to its p_self.validator (being a
           list or tuple o values).'''
        if self.translations:
            # Translations are available via a FieldTranslations instance
            language = language or o.guard.userLanguage
            r = self.translations.get('value', language, value=value)
        else:
            r = o.translate('%s_list_%s' % (self.labelId, value))
        return r

    def getFormattedValue(self, o, value, layout='view', showChanges=False,
                          language=None):
        '''Select-specific value formatting'''
        # Return an empty string if there is no p_value
        if Field.isEmptyValue(self, o, value) and not showChanges: return ''
        if isinstance(self.validator, Selection):
            # Value(s) come from a dynamic vocabulary
            val = self.validator
            if self.isMultiValued():
                r = [val.getText(o, v, self, language) for v in value]
            else:
                r = val.getText(o, value, self, language)
        else:
            # Values come from a fixed vocabulary whose texts are in i18n files
            _ = self.getTranslatedValue
            if self.isMultiValued():
                r = [_(o, v, language=language) for v in value]
            else:
                r = _(o, value, language=language)
        return r

    def getSearchValue(self, req, value=None):
        '''The search value as encoded in a search form for a Select field can
           be a simple string value or a list of values with an and/or
           operator.'''
        r = Field.getSearchValue(self, req, value=value)
        if isinstance(r, list):
            # It is a list of values. Get the operator to use.
            r = Operator.fromRequest(req, self.name)(*r)
        return r

    def getPossibleValues(self, o, withTranslations=False,
                          withBlankValue=False, blankLabel=None, className=None,
                          ignoreMasterValues=False):
        '''Returns the list of possible values for this field.
           If p_withTranslations is True, instead of returning a list of string
           values, the result is a list of tuples (s_value, s_translation).
           Moreover, p_withTranslations can hold a given language: in this case,
           this language is used instead of the user language. If
           p_withBlankValue is True, a blank value is prepended to the list,
           excepted if the type is multivalued. Used in combination with
           p_withTranslations being True, the i18n label for translating the
           blank value is given in p_blankLabel. If p_className is given, p_o is
           the tool and we are called from a context where no object is
           available (ie, a search form).'''
        # Get the user language for translations, from p_withTranslations
        lg = isinstance(withTranslations, str) and withTranslations or None
        req = o.req
        master = self.master
        if not ignoreMasterValues and master and callable(self.masterValue):
            # This field is an ajax-updatable slave. Get the master value...
            if master.valueIsInRequest(o, req):
                # ... from the request if available
                requestValue = master.getRequestValue(o)
                masterValues = master.getStorableValue(o, requestValue,
                                                       complete=True)
            elif not className:
                # ... or from the database if we are editing an object
                masterValues = master.getValue(o)
            else:
                # We don't have any master value
                masterValues = None
            # Get possible values by calling self.masterValue
            if masterValues:
                values = self.masterValue(o, masterValues)
            else:
                values = []
            # Manage parameter p_withTranslations
            if not withTranslations: res = values
            else:
                res = []
                for v in values:
                    # Translations may already have been included
                    if not isinstance(v ,tuple):
                        v = (v, self.getFormattedValue(o,v,language=lg))
                    res.append(v)
        else:
            # Get the possible values from attribute "[s]validator"
            validator = self.svalidator if className else self.validator
            if isinstance(validator, Selection):
                res = validator.getValues(o, className)
                if not withTranslations: res = [v[0] for v in res]
                elif isinstance(res, list): res = res[:]
            else:
                # The list of (static) values is in self.validator
                res = []
                for value in validator:
                    if withTranslations:
                        text = self.getTranslatedValue(o, value, language=lg)
                        res.append((value, text))
                    else:
                        res.append(value)
        if (withBlankValue == 'forced') or \
           (withBlankValue and not self.isMultiValued()):
            # Create the blank value to insert at the beginning of the list
            if withTranslations:
                label = blankLabel or self.noValueLabel
                blankValue = ('', o.translate(label, language=lg))
            else:
                blankValue = ''
            # Insert the blank value in the result
            if isinstance(res, tuple):
                res = (blankValue,) + res
            else:
                res.insert(0, blankValue)
        return res

    def validateValue(self, o, value):
        '''Ensure p_value is among possible values'''
        possibleValues = self.getPossibleValues(o, ignoreMasterValues=True)
        # In the context of a slave whose possible values depend on the selected
        # master, while the object is temp, possible values could not have been
        # computed yet.
        if not possibleValues: return
        if isinstance(value, str):
            error = value not in possibleValues
        else:
            error = False
            for v in value:
                if v not in possibleValues:
                    error = True
                    break
        if error: return o.translate('bad_select_value')

    def getStorableValue(self, o, value, complete=False):
        '''Get a multivalued value when appropriate'''
        if value and self.isMultiValued() and \
           (type(value) not in utils.sequenceTypes):
            value = [value]
        return value
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''An index for Rich fields, extracting pure text and splitting it into words'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.xml.extractor import Extractor
from appy.database.indexes.text import TextIndex

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class RichIndex(TextIndex):
    '''Index for Rich fields'''

    @classmethod
    def toIndexed(class_, value, field):
        '''Converts p_value, which is a chunk of XHTML code, into pure text and
           splits it into words.'''
        if not value: return
        # Use an XHTML text extractor. p_value represents a chunk of XHTML code,
        # but there is no guarantee that this code contains a single root tag,
        # so add one.
        extractor = Extractor(normalize=True, keepCRs=False)
        value = extractor.parse('<x>%s</x>' % value)
        # Tokenize the result, but text has already been normalized
        return TextIndex.toIndexed(value, field, normalize=False)

    @classmethod
    def toString(class_, o, value):
        '''Converts p_value to pure text via an XHTML text extractor'''
        # No need to normalize p_value here: it will be done by the global
        # "searchable" index.
        return Extractor(keepCRs=False).parse('<x>%s</x>' % value) if value \
                                                                   else None
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



#!/usr/bin/python

'''This script can be created on a Linux machine for creating a local Debian
(binary) repository.'''

import os, os.path
from appy.shared.utils import executeCommand

# Packages apache2 and dpkg-dev must be installed on the machine for enabling
# the Debian repository.
repoFolder = '/var/www/debianrepo'

# Create the repo folder if it does not exist
binaryFolder = os.path.join(repoFolder, 'binary')
if not os.path.exists(binaryFolder):
    os.makedirs(binaryFolder)

# Create the script that will allow to recompute indexes when packages are
# added or updated into the repository.
refreshScript = '''#!/bin/bash
cd %s
echo "(Re-)building indexes for binary packages..."
dpkg-scanpackages binary /dev/null | gzip -9c > binary/Packages.gz
echo "Done."
''' % repoFolder

curdir = os.getcwd()
os.chdir(repoFolder)
scriptName = os.path.join(repoFolder, 'refresh.sh')
if not os.path.exists(scriptName):
    f = file(scriptName, 'w')
    f.write(refreshScript)
    f.close()
executeCommand(['chmod', '-R', '755', repoFolder])
os.chdir(curdir)
print('Repository created.')



# ------------------------------------------------------------------------------
import re, os, os.path, base64, urllib, types

from appy import Object
from appy.px import Px
from appy.shared import utils as sutils

# Function for creating a Zope object ------------------------------------------
def createObject(folder, id, className, appName, wf=True, noSecurity=False,
                 initialComment='', initialState=None):
    '''Creates, in p_folder, object with some p_id. Object will be an instance
       of p_className from application p_appName. In a very special case (the
       creation of the config object), computing workflow-related info is not
       possible at this time. This is why this function can be called with
       p_wf=False.'''
    exec 'from Products.%s.%s import %s as ZopeClass' % \
         (appName, className, className)
    # Get the tool. It may not be present yet, maybe are we creating it now.
    if folder.meta_type.endswith('Folder'):
        # p_folder is a standard Zope (temp) folder
        tool = getattr(folder, 'config', None)
    else:
        # p_folder is an instance of a gen-class
        tool = folder.getTool()
    # Get the currently logged user
    user = None
    if tool: user = tool.getUser()
    # Checks whether the user may create this object if security is enabled
    if not noSecurity:
        klass = ZopeClass.wrapperClass.__bases__[-1]
        if not tool.userMayCreate(klass, checkInitiator=True):
            from AccessControl import Unauthorized
            raise Unauthorized("User can't create instances of %s" % \
                               klass.__name__)
    # Prevent object creation if "id" refers to an existing object
    if tool and tool.getObject(id, brain=True):
        raise Exception('object with id "%s" already exist.' % id)
    # Create the object
    obj = ZopeClass(id)
    folder._objects = folder._objects + ({'id':id, 'meta_type':className},)
    folder._setOb(id, obj)
    obj = folder._getOb(id) # Important. Else, obj is not really in the folder.
    obj.portal_type = className
    obj.id = id
    # If no user object is there, we are at startup, before default User
    # instances are created.
    userId = user and user.login or 'system'
    obj.creator = userId
    from DateTime import DateTime
    obj.created = DateTime()
    obj.modified = obj.created
    from persistent.mapping import PersistentMapping
    obj.__ac_local_roles__ = PersistentMapping({ userId: ['Owner'] })
    if wf: obj.initializeWorkflow(initialComment, initialState)
    return obj

# ------------------------------------------------------------------------------
upperLetter = re.compile('[A-Z]')
def produceNiceMessage(msg):
    '''Transforms p_msg into a nice msg.'''
    res = ''
    if msg:
        res = msg[0].upper()
        for c in msg[1:]:
            if c == '_':
                res += ' '
            elif upperLetter.match(c):
                res += ' ' + c.lower()
            else:
                res += c
    return res

# ------------------------------------------------------------------------------
class SomeObjects:
    '''Represents a bunch of objects retrieved from a reference or a query in
       the catalog.'''
    DEAD_BRAIN = 'brain removed for missing object@%s.'

    def __init__(self, objects=None, batchSize=None, startNumber=0,
                 noSecurity=False):
        # The objects (more precisely, p_objects start being brains)
        self.objects = objects or []
        # self.objects may only represent a part of all available objects
        self.totalNumber = len(self.objects)
        # self.objects' max length
        self.batchSize = batchSize or self.totalNumber
        # The index of first object in self.objects in the whole list
        self.startNumber = startNumber
        self.noSecurity = noSecurity
        # Some brains may correspond to unexistent objects
        self.deadBrains = None

    def addDeadBrain(self, brain):
        '''Adds p_brain to the list of dead brains. More precisely, p_brain is
           not a brain but a brain's path.'''
        dead = self.deadBrains
        if dead:
            dead.append(brain)
        else:
            self.deadBrains = [brain]

    def removeDeadBrains(self, tool):
        '''Removes dead brains from the catalog'''
        catalog = tool.getPath('/catalog')
        for path in self.deadBrains:
            catalog.uncatalog_object(path)
            tool.log(self.DEAD_BRAIN % path, noUser=True, type='error')
            self.totalNumber -= 1

    def brainsToObjects(self, tool):
        '''self.objects has been populated from catalog's brains, not from true
           objects. This method turns them (or some of them depending on
           batchSize and startNumber) into real objects.

           If self.noSecurity is True, it gets the objects even if the logged
           user does not have the right to get them.'''
        # Browse (a sub-set of) brains
        start = self.startNumber
        total = self.totalNumber
        i = start
        end = start + self.batchSize
        # Prepare data structures
        brains = self.objects
        r = []
        noSecurity = self.noSecurity
        while (i < end) and (i < total):
            brain = brains[i]
            obj = tool.toObject(brain, noSecurity=True)
            if obj is None:
                # A dead brain. Add it to the list: we will remove it
                # afterwards. If we remove dead brains while browsing brains, it
                # produces problems within the list of brains and all brains are
                # not walked.
                self.addDeadBrain(brain.getPath())
            else:
                r.append(obj)
            i += 1
        # Replace brains with (found) objects
        self.objects = r
        if self.deadBrains: self.removeDeadBrains(tool)

# ------------------------------------------------------------------------------
def splitIntoWords(text, ignore=2, ignoreNumbers=False, words=None):
    '''Splits p_text into words. If p_words is None, it returns the set of
       words (no duplicates). Else, it adds one entry in the p_words dict for
       every encountered word.

       Words whose length is <= p_ignore are ignored, excepted, if
       p_ignoreNumbers is False, words being numbers.'''
    # Split p_text into words
    r = text.split()
    # Browse words in reverse order and remove shorter ones
    i = len(r) - 1
    keepIt = None
    while i > -1:
        word = r[i]
        # Keep this word or not ?
        if len(word) <= ignore:
            keepIt = not ignoreNumbers and word.isdigit()
        else:
            keepIt = True
        # Update "r" or "words" accordingly
        if words != None:
            # Add the word to "words" when we must keep it
            if keepIt:
                words[word] = None
        else:
            # Remove the word from "r" if we must not keep it
            if not keepIt:
                del r[i]
        i -= 1
    # Return the result as a set when relevant
    if words == None:
        return set(r)

# ------------------------------------------------------------------------------
class Keywords:
    '''This class allows to handle keywords that a user enters and that will be
       used as basis for performing requests in a TextIndex/XhtmlIndex.'''
    toRemove = '?-+*()'
    def __init__(self, keywords, operator='AND', ignore=2):
        # Clean the p_keywords that the user has entered
        words = sutils.normalizeText(keywords)
        if words == '*': words = ''
        for c in self.toRemove: words = words.replace(c, ' ')
        self.keywords = splitIntoWords(words, ignore=ignore)
        # Store the operator to apply to the keywords (AND or OR)
        self.operator = operator

    def merge(self, other, append=False):
        '''Merges our keywords with those from p_other. If p_append is True,
           p_other keywords are appended at the end; else, keywords are appended
           at the begin.'''
        for word in other.keywords:
            if word not in self.keywords:
                if append:
                    self.keywords.append(word)
                else:
                    self.keywords.insert(0, word)

    def get(self):
        '''Returns the keywords as needed by the TextIndex.'''
        if self.keywords:
            op = ' %s ' % self.operator
            return op.join(self.keywords)+'*'
        return ''

# ------------------------------------------------------------------------------
def getClassName(klass, appName=None):
    '''Generates, from appy-class p_klass, the name of the corresponding
       Zope class. For some classes, name p_appName is required: it is
       part of the class name.'''
    moduleName = klass.__module__
    if (moduleName == 'appy.gen.model') or moduleName.endswith('.wrappers'):
        # This is a model (generation time or run time)
        res = appName + klass.__name__
    elif klass.__bases__ and (klass.__bases__[-1].__module__=='appy.gen.utils'):
        # This is a customized class (inherits from appy.gen.Tool, User,...)
        res = appName + klass.__bases__[-1].__name__
    else: # This is a standard class
        res = klass.__module__.replace('.', '_') + '_' + klass.__name__
    return res

# ------------------------------------------------------------------------------
def callMethod(obj, method, klass=None, cache=True):
    '''This function is used to call a p_method on some Appy p_obj. m_method
       can be an instance method on p_obj; it can also be a static method. In
       this latter case, p_obj is the tool and the static method, defined in
       p_klass, will be called with the tool as unique arg.

       A method cache is implemented on the request object (available at
       p_obj.request). So while handling a single request from the ui, every
       method is called only once. Some method calls must not be cached (ie,
       values of Computed fields). In this case, p_cache will be False.'''
    rq = obj.request
    # Disable cache if the request is fake. Indeed, caching is interesting when
    # we are managing a real ui request.
    if rq.__class__.__name__ == 'Object':
        cache = False
    else:
        # Create the method cache if it does not exist on the request
        if not hasattr(rq, 'methodCache'): rq.methodCache = {}
    # If m_method is a static method or an instance method, unwrap the true
    # Python function object behind it.
    methodType = method.__class__.__name__
    if methodType == 'staticmethod':
        method = method.__get__(klass)
    elif (methodType == 'instancemethod') and \
         (type(method.im_self) != types.ClassType):
        # An instance method. If im_self is a class, it is a classmethod.
        method = method.im_func
    # Disable caching for lambda functions
    funName = method.func_name
    if funName == '<lambda>': cache = False
    # Call the method if cache is not needed
    if not cache: return method(obj)
    # If first arg of method is named "tool" instead of the traditional "self",
    # we cheat and will call the method with the tool as first arg. This will
    # allow to consider this method as if it was a static method on the tool.
    # Every method call, even on different instances, will be cached in a unique
    # key.
    cheat = False
    if not klass and (method.func_code.co_varnames[0] == 'tool'):
        prefix = obj.klass.__name__
        obj = obj.tool
        cheat = True
    # Build the key of this method call in the cache.
    # First part of the key: the p_obj's uid (if p_method is an instance method)
    # or p_className (if p_method is a static method).
    if not cheat:
        if klass:
            prefix = klass.__name__
        else:
            prefix = obj.id
    # Second part of the key: p_method name
    key = '%s:%s' % (prefix, funName)
    # Return the cached value if present in the method cache
    if key in rq.methodCache:
        return rq.methodCache[key]
    # No cached value: call the method, cache the result and return it
    res = method(obj)
    rq.methodCache[key] = res
    return res

# Functions for manipulating the authentication cookie -------------------------
def readCookie(request, onResponse=False):
    '''Returns the tuple (login, password, ctx) read from the authentication
       cookie received in p_request. If no user is logged, its returns
       (None, None, None).'''
    if not onResponse:
        cookie = request.get('_appy_', None)
    else:
        cookie = request.RESPONSE.cookies['_appy_']['value']
    if not cookie: return None, None, None
    cookieValue = base64.decodestring(urllib.unquote(cookie))
    if ':' not in cookieValue: return None, None, None
    # Extract the context from the cookieValue
    r, context = cookieValue.rsplit(':', 1)
    # Maintain compatibility with old-style, context-free cookies
    if ':' not in r:
        r = cookieValue
        context = None
    # Extract login and password
    login, password = r.split(':', 1)
    return login, password, context

def writeCookie(login, password, ctx, request):
    '''Encode p_login, p_password and p_ctx into a cookie'''
    res = base64.encodestring('%s:%s:%s' % (login,password,ctx or '')).rstrip()
    request.RESPONSE.setCookie('_appy_', urllib.quote(res), path='/')

def updateCookie(request, ctx, onResponse=False):
    '''Updates the authentication context within the Appy cookie. If
       p_onResponse is True, we get the cookie from the response object instead
       of the request object.'''
    login, password, oldCtx = readCookie(request, onResponse=onResponse)
    if login is None:
        # There is no cookie. When SSO is enabled, cookies are normally not
        # used, excepted if an authentication context comes into play. In that
        # case, the cookie is only used to store this context. In order to
        # represent this fact, set, in the cookie, the user login as "_s_s_o_".
        # Such a cookie means: "this is a special cookie, only used to store the
        # authentication context. Check SSO to get the user login".
        login = '_s_s_o_'
    writeCookie(login, password, ctx, request)

# ------------------------------------------------------------------------------
def initMasterValue(v):
    '''Standardizes p_v as a list of strings, excepted if p_v is a method'''
    if callable(v): return v
    if not isinstance(v, bool) and not v: res = []
    elif type(v) not in sutils.sequenceTypes: res = [v]
    else: res = v
    return [str(v) for v in res]

# ------------------------------------------------------------------------------
class No:
    '''When you write a workflow condition method and you want to return False
       but you want to give to the user some explanations about why a transition
       can't be triggered, do not return False, return an instance of No
       instead. When creating such an instance, you can specify an error
       message.'''
    def __init__(self, msg): self.msg = msg
    def __nonzero__(self): return False
    def __repr__(self): return '<No: %s>' % self.msg

class MessageException(Exception): pass

# ------------------------------------------------------------------------------
class Model: pass
class Tool(Model):
    '''Subclass me to extend or modify the Tool class'''
class User(Model):
    '''Subclass me to extend or modify the User class'''

# ------------------------------------------------------------------------------
class Collapsible:
    '''Represents a chunk of HTML code that can be collapsed/expanded via
       clickable icons.'''
    # Various sets of icons can be used. Each one has a CSS class in appy.css
    iconSets = {'expandCollapse': Object(expand='expand', collapse='collapse'),
                'showHide': Object(expand='show', collapse='hide'),
                'showHideInv': Object(expand='hide', collapse='show')}

    # Icon allowing to collapse/expand a chunk of HTML
    px = Px('''
     <img var="coll=collapse; icons=coll.icons"
          id=":'%s_img' % coll.id" align=":coll.align" class=":coll.css"
          onclick=":'toggleCookie(%s,%s,%s,%s,%s)' % (q(coll.id), \
                    q(coll.display), q(coll.default), \
                    q(icons.expand), q(icons.collapse))"
       src=":coll.expanded and url(icons.collapse) or url(icons.expand)"/>''')

    def __init__(self, id, request, default='collapsed', display='block',
                 icons='expandCollapse', align='left'):
        '''p_display is the value of style attribute "display" for the XHTML
           element when it must be displayed. By default it is "block"; for a
           table it must be "table", etc.'''
        self.id = id # The ID of the collapsible HTML element
        self.request = request # The request object
        self.default = default
        self.display = display
        self.align = align
        # Must the element be collapsed or expanded ?
        self.expanded = request.get(id, default) == 'expanded'
        self.style = 'display:%s' % (self.expanded and self.display or 'none')
        # The name of the CSS class depends on the set of applied icons
        self.css = icons
        self.icons = self.iconSets[icons]

# ------------------------------------------------------------------------------
class LinkTarget:
    '''Represents information about the target of an HTML "a" tag'''

    def __init__(self, klass=None, back=None, forcePopup=False):
        '''The HTML "a" tag must lead to a page for viewing or editing an
           instance of some p_klass. If this page must be opened in a popup
           (depends on attribute p_klass.popup), and if p_back is specified,
           when coming back from the popup, we will ajax-refresh a DOM node
           whose ID is specified in p_back.'''
        # The link leads to a instance of some p_klass
        self.klass = klass
        # Does the link lead to a popup ?
        if forcePopup:
            toPopup = True
        else:
            toPopup = klass and hasattr(klass, 'popup')
        # Determine the target of the "a" tag
        self.target = toPopup and 'appyIFrame' or '_self'
        # If the link leads to a popup, a "onClick" attribute must contain the
        # JS code that opens the popup.
        if toPopup:
            # Create the chunk of JS code to open the popup
            size = getattr(klass, 'popup', '350px')
            if isinstance(size, basestring):
                params = "%s,null" % size[:-2] # Width only
            else: # Width and height
                params = "%s,%s" % (size[0][:-2], size[1][:-2])
            # If p_back is specified, included it in the JS call
            if back: params += ",'%s'" % back
            self.onClick = "openPopup('iframePopup',null,%s)" % params
        else:
            self.onClick = ''

    def getOnClick(self, back):
        '''Gets the "onClick" attribute, taking into account p_back DOM node ID
           that was unknown at the time the LinkTarget instance was created.'''
        # If we must not come back from a popup, return an empty string
        r = self.onClick
        if not r: return r
        return r[:-1] + ",'%s')" % back

    def __repr__(self):
        return '<LinkTarget for=%s,target=%s,onClick=%s>' % \
               (self.klass.__name__, self.target, self.onClick or '-')
# ------------------------------------------------------------------------------



'''Extracts statistics from an ODF document (meta.xml)'''

# ------------------------------------------------------------------------------
from appy.shared.zip import unzip
from appy.shared.utils import getOsTempFolder
from appy.shared.xml_parser import XmlUnmarshaller

# ------------------------------------------------------------------------------
class OdfMetadata:
    '''Represents metadata about an ODF document'''
    def __init__(self, metadata):
        '''Initialises instance attributes based on m_metadata that contains the
           unmarshalled content of some ODF file's meta.xml.'''
        # Store document statistics
        stats = metadata.get('document-statistic')
        for name, value in stats.__dict__.iteritems():
            # Value should be integer
            try:
                value = int(value)
            except ValueError:
                pass
            # Attribute names can contain dashes
            setattr(self, name.replace('-', ''), value)

# ------------------------------------------------------------------------------
class MetadataReader:
    metaTagTypes = {'meta:document-statistic': 'object'}

    def __init__(self, fileName):
        self.fileName = fileName

    def run(self):
        '''Extracts metadata and statistics from the ODF file whose name is in
           self.fileName and returns an OdfMetadata instance.'''
        # Unzip the ODF file into a temp folder
        tempFolder = getOsTempFolder(sub=True)
        metaXml = unzip(self.fileName, tempFolder, odf=True)['meta.xml']
        parser = XmlUnmarshaller(tagTypes=self.metaTagTypes)
        return OdfMetadata(parser.unmarshall(metaXml).meta)
# ------------------------------------------------------------------------------



'''The Appy meta-model contains meta-classes representing classes of Appy
   classes: essentially, Appy classes and workflows.'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.model import Model

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
ATTR       = 'Attribute "%s" in %s "%s"'
UNDERS     = 'contains at least one underscore, which is not allowed'

US_IN_ATTR = '%s %s. Please consider naming your fields, searches, states ' \
             'and transitions in camelCase style, the first letter being a ' \
             'lowercase letter. For example: "myField", instead of ' \
             '"my_field". Furthermore, there is no need to define "private" ' \
             'fields starting with an underscore.' % (ATTR, UNDERS)
US_IN_CLS = 'ClassS or workflow "%%s" %s. Please consider naming your classes ' \
            'and workflows in CamelCase style, the first letter being an ' \
            'uppercase letter. For example: "MyClass", instead of ' \
            '"my_class".' % UNDERS
UP_NAME    = '%s must start with a lowercase letter. This rule holds for any ' \
             'field, search, state or transition.' % ATTR
LOW_NAME   = 'Name of %s "%s" must start with an uppercase letter.'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Meta:
    '''Abstract base class representing a Appy class or workflow'''

    def __init__(self, class_, appOnly):
        # p_class_ is the Python class found in the Appy app. Its full name is
        #              <Python file name>.py.<class name>
        self.python = class_
        # Make this link bidirectional
        class_.meta = self
        # Its name. Ensure it is valid.
        self.name = self.checkClassName(class_)
        # If p_appOnly is True, stuff related to the Appy base model must not be
        # defined on this meta-class. This minimalist mode is in use when
        # loading the model for creating or updating translation files.
        self.appOnly = appOnly

    def asString(self):
        r = '<class %s.%s' % (self.__module__, self.name)
        for attribute in self.attributes.keys():
            r += '\n %s:' % attribute
            for name, field in getattr(self, attribute).items():
                r += '\n  %s : %s' % (name, str(field))
        return r

    def checkClassName(self, class_):
        '''The name of a class or workflow must start with an uppercase letter
           and cannot contain an underscore.'''
        name = class_.__name__
        if '_' in name:
            raise Model.Error(US_IN_CLS % (name))
        if name[0].islower():
            type = self.__class__.__name__.lower()
            raise Model.Error(LOW_NAME % (type, name))
        return name

    def checkAttributeName(self, name):
        '''Checks if p_name is valid for being used as attribute (field, search,
           state or transition) for this class or workflow.'''
        # Underscores are not allowed
        if '_' in name:
            type = self.__class__.__name__.lower()
            raise Model.Error(US_IN_ATTR % (name, type, self.name))
        # The name must start with a lowercase char
        if not name[0].islower():
            type = self.__class__.__name__.lower()
            raise Model.Error(UP_NAME % (name, type, self.name))

    def __repr__(self): return '<class %s.%s>' % (self.__module__, self.name)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



import os.path
import appy

def getAppyPath():
    return os.path.dirname(appy.__file__)

xhtmlChunk = '''
<img alt="" src="./images/linux.jpg"
     style="height:100px; width:100px" title="Image"/>
<img alt="" src="./images/linux.jpg"
     style="height:200px; width:100px; float:right;" title="Image"/>
<!-- Image with no defined width or height -->
<img src="./images/imio.png"/>
<!-- Image with html attributes "width" and "height" -->
<img src="./images/imio.png" width="74px" height="145px"/>
<p></p>
<table>
 <tbody><tr><td><p><img alt="" src="./images/linux.jpg"/></p></td></tr></tbody>
</table>
<img/> <!-- This will be ignored because having no "src" attr -->
'''



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.px import Px
from appy.model.fields import Field, Show
from appy.ui.layout import Layouts, Layout
from appy.model.searches import Search, UiSearch

# Error messages - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
UNFREEZ   = 'This field is unfreezable.'
METHOD_NO = 'Specify a method in parameter "method".'
METHOD_KO = 'Wrong value "%s". Parameter "method" must contain a method or ' \
            'a PX.'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Computed(Field):
    '''Useful for computing a custom field via a Python method'''

    class Layouts(Layouts):
        '''Computed-specific layouts'''
        # Layouts for fields in a grid group, with description
        gd = Layouts('f-drvl')
        # Idem, but with a help icon
        gdh = Layouts('f-dhrvl')

    view = cell = edit = Px('''<x if="field.plainText">:value</x><x
      if="not field.plainText">::value</x>''')

    search = Px('''
     <input type="text" name=":widgetName" maxlength=":field.maxChars"
            size=":field.width" value=":field.sdefault"/>''')

    def __init__(self, multiplicity=(0,1), default=None, defaultOnEdit=None,
      show=None, page='main', group=None, layouts=None, move=0, indexed=False,
      mustIndex=True, indexType=None, indexValue=None, emptyIndexValue=None,
      searchable=False, filterField=None, readPermission='read',
      writePermission='write', width=None, height=None, maxChars=None,
      colspan=1, method=None, formatMethod=None, plainText=False, master=None,
      masterValue=None, focus=False, historized=False, mapping=None,
      generateLabel=None, label=None, sdefault='', scolspan=1, swidth=None,
      sheight=None, context=None, view=None, cell=None, edit=None, xml=None,
      translations=None, unfreezable=False, validable=False):
        # The Python method used for computing the field value, or a PX
        self.method = method
        # A specific method for producing the formatted value of this field.
        # This way, if, for example, the value is a DateTime instance which is
        # indexed, you can specify in m_formatMethod the way to format it in
        # the user interface while m_method computes the value stored in the
        # catalog.
        self.formatMethod = formatMethod
        # Does field computation produce plain text or XHTML?
        self.plainText = plainText
        if isinstance(method, Px):
            # When field computation is done with a PX, the result is XHTML
            self.plainText = False
        # Determine default value for "show"
        if show is None:
            # XHTML content in a Computed field generally corresponds to some
            # custom XHTML widget. This is why, by default, we do not render it
            # in the xml layout.
            show = Show.E_ if self.plainText else Show.TR
        # If method is a PX, its context can be given in p_context
        self.context = context
        # For any other Field subclass, the type of the index to use is
        # statically determined. But for an indexed Computed field, which may
        # hold any data of any type, the index type must be specified in
        # p_indexType. Choose it in this list, depending on the values that your
        # field will store.
        self.indexType = indexType or 'Index'
        # ----------------------------------------------------------------------
        #   Index type   | Suitable...
        # ----------------------------------------------------------------------
        #     "Index"    | ... in most cases, when none of the following index
        #                |     types are appropriate ;
        # ----------------------------------------------------------------------
        #   "RefIndex"   | ... for storing lists of Appy objects (=instances of
        #                |     Appy classes). You should not use this type of 
        #                |     index and use a Ref field for storing lists of
        #                |     objects ;
        # ----------------------------------------------------------------------
        #   "DateIndex"  | ... if your field stores DateTime instances ;
        # ----------------------------------------------------------------------
        #   "TextIndex"  | ... if your field stores raw text and you want to
        #                |     index words in found in it ;
        # ----------------------------------------------------------------------
        #   "RichIndex"  | ... if your field stores a chunk of XHTML code and 
        #                |     you want to index text extracted from it ;
        # ----------------------------------------------------------------------
        #  "FloatIndex"  | ... if your field stores float values ;
        # ----------------------------------------------------------------------
        # "BooleanIndex" | ... if your field stores boolean values.
        # ----------------------------------------------------------------------
        # Call the base constructor
        Field.__init__(self, None, multiplicity, default, defaultOnEdit, show,
          page, group, layouts, move, indexed, mustIndex, indexValue,
          emptyIndexValue, searchable, filterField, readPermission,
          writePermission, width, height, None, colspan, master, masterValue,
          focus, historized, mapping, generateLabel, label, sdefault, scolspan,
          swidth, sheight, False, False, view, cell, edit, xml, translations)
        # When a custom widget is built from a computed field, its values are
        # potentially editable and validable, so "validable" must be True.
        self.validable = validable
        # One classic use case for a Computed field is to build a custom widget.
        # In this case, self.method stores a PX or method that produces, on
        # view or edit, the custom widget. Logically, you will need to store a
        # custom data structure on the object, in an attribute named according
        # to this field, ie o.[self.name]. Typically, you will set or update a
        # value for this attribute in m_onEdit, by getting, on the o.req object,
        # values encoded by the user in your custom widget (edit mode). This
        # "custom widget" use case is incompatible with "freezing". Indeed,
        # freezing a Computed field implies storing the computed value at
        # o.[self.name] instead of recomputing it as usual. So if you want to
        # build a custom widget, specify the field as being unfreezable.
        self.unfreezable = unfreezable
        # Set a filter PX if this field is indexed with a TextIndex
        if self.indexed and (self.indexType == 'TextIndex'):
            self.filterPx = 'pxFilterText'
        self.checkParameters()

    def checkParameters(self):
        '''Ensures a valid method is specified'''
        method = self.method
        # A method must be there
        if not method: raise Exception(METHOD_NO)
        # It cannot be a string, but a true method
        if isinstance(method, str): raise Exception(METHOD_KO % method)

    def renderPx(self, o, px):
        '''Renders the p_px and returns the result'''
        traversal = o.traversal
        context = traversal.context or traversal.createContext()
        # Complete the context when relevant
        custom = self.context
        custom = custom if not callable(custom) else custom(o)
        if custom:
            context.update(custom)
        return px(context)

    def renderSearch(self, o, search):
        '''Executes the p_search and return the result'''
        req = o.req
        # This will allow the UI to find this search
        req.search = '%d,%s,view' % (o.iid, self.name)
        req.className = search.container.name
        traversal = o.traversal
        existingContext = traversal.context
        context = traversal.createContext()
        r = context.uiSearch.pxResult(context)
        # Reinitialise the context correctly
        if existingContext:
            traversal.context = existingContext
        return r

    def getSearch(self, o):
        '''Gets the Search instance possibly linked to this Computed field'''
        method = self.method
        if not method: return
        if isinstance(method, Search): return method
        # Maybe a dynamically-computed Search ?
        r = self.callMethod(o, method, cache=False)
        if isinstance(r, Search): return r

    def getValue(self, o, name=None, layout=None, single=None,
                 forceCompute=False):
        '''Computes the field value on p_obj or get it from the database if it
           has been frozen.'''
        # Is there a database value ?
        if not self.unfreezable and not forceCompute:
            r = o.values.get(self.name)
            if r is not None: return r
        # Compute the value
        meth = self.method
        if not meth: return
        if isinstance(meth, Px): return self.renderPx(o, meth)
        elif isinstance(meth, Search): return self.renderSearch(o, meth)
        else:
            # self.method is a method that will return the field value
            r = self.callMethod(o, meth, cache=False)
            # The field value can be a dynamically computed PX or Search
            if isinstance(r, Px): return self.renderPx(o, r)
            elif isinstance(r, Search): return self.renderSearch(o, r)
            return r

    def getFormattedValue(self, o, value, layout='view', showChanges=False,
                          language=None):
        if self.formatMethod:
            r = self.formatMethod(o, value)
        else:
            r = value
        if not isinstance(r, str): r = str(r)
        return r

    # If you build a custom widget with a Computed field, Appy can't tell if the
    # value in your widget is complete or not. So it returns True by default.
    # It is up to you, in method obj.validate, to perform a complete validation,
    # including verifying if there is a value if your field is required.
    def isCompleteValue(self, o, value): return True

    def freeze(self, o, value=None):
        '''Normally, no field value is stored for a Computed field: the value is
           computed on-the-fly by p_self.method. But if you freeze it, a value
           is stored: either p_value if not None, or the result of calling
           p_self.method else. Once a Computed field value has been frozen,
           everytime its value will be requested, the frozen value will be
           returned and p_self.method will not be called anymore. Note that the
           frozen value can be unfrozen (see method below).'''
        if self.unfreezable: raise Exception(UNFREEZ)
        # Compute for the last time the field value if p_value is None
        if value is None: value = self.getValue(o, forceCompute=True)
        # Freeze the given or computed value (if not None) in the database
        if value is not None: o.values[self.name] = value

    def unfreeze(self, o):
        '''Removes the database value that was frozen for this field on p_o'''
        if self.unfreezable: raise Exception(UNFREEZ)
        if self.name in o.values: del(o.values[self.name])
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



#!/usr/bin/python3

# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import os, shutil
from pathlib import Path

import appy
from appy.tr.po import File
from appy.bin import Program
from appy.utils.path import chown
from appy.utils.loc import Counter
from appy.tr.updater import Updater
from appy.deploy import parseSpecifier
from appy.deploy.subversion import Subversion

# File templates - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# <site>/bin/site
site = """
#!/usr/bin/python3

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import sys
from pathlib import Path

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
site = Path('{self.folder}')
app  = Path('{self.app}')
ext  = '{self.ext}'
ext  = Path(ext) if ext else None

paths = [site, site/'lib', app.parent]
if ext: paths.append(ext.parent)

for path in paths:
    path = str(path)
    if path not in sys.path:
        sys.path.insert(0, path)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Import Appy after sys.path has been updated
from appy.bin.run import Run

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
if __name__ == '__main__':
    Run().run(site, app, ext)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
"""

# <site>/bin/deploy
deploy = """
#!/usr/bin/python3
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import sys
from appy.bin.deploy import Deploy

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
if __name__ == '__main__':
    sys.argv.insert(1, '{self.folder}')
    sys.argv.insert(1, '{self.app}')
    Deploy().run()
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
"""

# <site>/config.py
config = """
# -*- coding: utf-8 -*-

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
app = '{self.app}'
site = '{self.folder}'

# Complete the config  - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
def complete(c):
    c.model.set(app)
    c.server.set(app, site)
    c.server.port = {self.port}
    c.database.set(site + '/var')
    c.log.set(site + '/var/site.log', site + '/var/app.log')
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
"""

# <app>/__init__.py
appInit = '''
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.database import log
from appy.server import guard, scheduler, backup
from appy import database, model, ui, server, deploy, peer

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import appy
class Config(appy.Config):
    server = server.Config()
    security = guard.Config()
    backup = backup.Config()
    jobs = scheduler.Config()
    database = database.Config()
    log = log.Config()
    model = model.Config()
    ui = ui.Config()
    deploy = deploy.Config()
    peers = peer.Config()

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
try: # to import the site-specific configuration file
    import config
    config.complete(Config)
except ImportError:
    # The "config" module may not be present at maketime
    pass
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
'''

# <app>/make
appMake = '''
#!/usr/bin/python3
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from pathlib import Path
from appy.bin.make import App

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
App(Path('.')).update()
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
'''

# <app>/static/robots.txt
appRobots = '''
# More information about this file on http://robots-txt.com
User-agent: *
Disallow: /
'''

# <ext>/__init__.py
extInit = '''
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Import the app's main Config instance
from %s import Config as c

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Declare this extension
from pathlib import Path
c.declareExt(Path('%s'))

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Apply ext's changes to this instance
# c.<attribute> = ...

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Target:
    '''Abstract class representing the artefact to create or update'''

    def createFile(self, path, content, permissions=None):
        '''Create a file on disk @ p_path, with some p_content and
           p_permissions.'''
        folder = path.parent
        # Create the parent folder if it does not exist
        if not folder.exists(): folder.mkdir(parents=True)
        # Patch content with variables
        content = content.format(self=self)[1:]
        # Create the file
        path = str(path)
        f = open(path, 'w')
        f.write(content)
        f.close()
        # Set specific permissions on the created file when required
        if permissions: os.chmod(path, permissions)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Site(Target):
    '''A site listens to some port and runs an app'''

    # A site is made of 3 sub-folders:
    folders = ('bin', 'lib', 'var')
    # "bin" stores the scripts for controlling the app (start/stop), backuping
    #       or restoring its data, etc;
    # "lib" stores or links the Python modules available to the site. It
    #       contains at least the app;
    # "var" stores the data produced or managed by the site: database files and
    #       folders and log files.

    def __init__(self, folder, app, port, owner, dependencies, ext):
        # The base folder for the site
        self.folder = folder
        # The folder or distant specifier for the app and its ext
        self.app = app
        self.ext = ext
        # The port on which the site will listen
        self.port = port
        # The owner of the site tree (if None, the user executing the script
        # will become owner of the tree).
        if owner:
            if ':' in owner:
                self.owner, self.ownerGroup = owner.split(':')
            else:
                self.owner = owner
                self.ownerGroup = None
        else:
            self.owner = self.ownerGroup = None
        # Site dependencies
        self.dependencies = dependencies

    def create(self):
        '''Creates a fresh site'''
        # Create the root folder (and parents if they do not exist)
        self.folder.mkdir(parents=True)
        # Create the sub-folders
        for name in Site.folders: (self.folder/name).mkdir()
        # If the app or ext are distant, get, in lib/<appOrExtName>, a copy from
        # the server.
        lib = self.folder / 'lib'
        if not isinstance(self.app, Path):
            self.app = Subversion(self.app).checkout(lib)
        ext = self.ext
        if ext:
            self.ext = ext if isinstance(ext, Path) \
                           else Subversion(ext).checkout(lib)
        else:
            self.ext = ''
        # Integrate dependencies into the site
        dependencies = self.dependencies
        if dependencies:
            for dep in dependencies:
                if isinstance(dep, Path):
                    # A dependency to a local package. Symlink it in <site>/lib.
                    os.symlink(dep, lib / dep.name)
                else:
                    # A specifier. Checkout a copy in <site>/lib.
                    Subversion(dep).checkout(lib)
        # Create <site>/bin/site and <site>/bin/deploy
        self.createFile(self.folder/'bin'/'site', site, 0o770)
        self.createFile(self.folder/'bin'/'deploy', deploy, 0o770)
        # Create <site>/config.py
        self.createFile(self.folder/'config.py', config)
        # Chown the tree to p_self.owner if it has been specified
        owner = self.owner
        if owner:
            chown(self.folder, owner, group=self.ownerGroup, recursive=True)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class App(Target):
    '''An app is a piece of software being Appy-compliant'''

    # Mandatory app sub-folders
    subFolders = (
      'tr',    # Stores "po" and "pot" files (for translations)
      'static' # Stores any static content (images, CSS or Javascript files...)
    )

    # Mandatory files to have in <appFolder>/static
    files = (
      # The icon retrieved by Safari
      'appleicon.png',
      # he icon retrieved by other browsers
      'favicon.ico',
      # The file read by robots
      'robots.txt'
    )

    def __init__(self, folder):
        # The base folder for the app. Ensure it is absolute.
        if not folder.is_absolute():
            folder = folder.resolve()
        self.folder = folder

    def create(self):
        '''Creates a new app'''
        # Create <app>/__init__.py
        path = self.folder / '__init__.py'
        if not path.exists():
            self.createFile(path, appInit)
        # Create <app>/make
        path = self.folder / 'make'
        if not path.exists():
            self.createFile(path, appMake, 0o770)
        # Create base folders
        for name in App.subFolders:
            folder = self.folder / name
            if not folder.exists(): folder.mkdir()
            if name == 'static':
                # Ensure mandatory files are present and create them if not
                robots = folder / 'robots.txt' # The file read by robots
                if not robots.exists():
                    self.createFile(robots, appRobots)
                # The icon retrieved by browsers, shown besides the address bar
                for name in ('favicon.ico', 'appleicon.png'):
                    icon = folder / name
                    if not icon.exists():
                        # Copy the default one in Appy
                        ico = '%s/ui/static/%s' % (str(appy.path), name)
                        shutil.copy(ico, str(icon))

    def update(self):
        '''Updates an existing app'''
        print('Updating %s...' % self.folder)
        # Ensure the base files and folders are created
        self.create()
        # Call the translations updater that will create or update translation
        # files for this app.
        Updater(self.folder / 'tr').run()
        # Count the lines of code in the app
        Counter(self.folder).run()
        print('Done.')

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Ext(App):
    '''An ext is an extension to a given app'''

    def __init__(self, folder, app):
        # Call the base constructor
        App.__init__(self, folder)
        # The app's Path
        self.app = app

    def getAppLanguages(self):
        '''Retrieve the languages supported by the app'''
        # It is done by searching the app's .po files. Doing it by importing the
        # app and reading its Config instance, in order to access attribute
        # Config.languages, has been felt more risky, because importing the app
        # could lead to import errors.
        r = []
        name = self.app.name
        for po in (self.app / 'tr').glob('%s-*.po' % self.app.name):
            r.append(po.stem.rsplit('-')[1])
        return r

    def create(self):
        '''Creates a new ext'''
        # Create <ext>/__init__.py
        path = self.folder / '__init__.py'
        if not path.exists():
            self.createFile(path, extInit % (self.app.name, str(self.folder)))
        # Create base folders
        for name in App.subFolders:
            folder = self.folder / name
            if not folder.exists(): folder.mkdir()
        # Create a "po" file for every app language, in case the ext needs to
        # override some i18n labels.
        for lang in self.getAppLanguages():
            # Create an empty file named <ext>/tr/Custom-<lang>.po
            name = 'Custom-%s.po' % lang
            File(self.folder / 'tr' / name).generate()

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Make(Program):
    '''This program allows to create or update an app or a site'''
    # We can create/update the following artefacts
    allowedTargets = ('app', 'site', 'ext')

    # Help messages
    APP_S        = 'url=<url>, [name=<name>,login=<login>, password=<password>]'
    HELP_TARGET  = 'can be "app" for creating/updating an app, "site" for ' \
                   'creating a site or "ext" for creating an extension to an ' \
                   'app.'
    HELP_FOLDER  = 'folder is the absolute path to the base folder for the ' \
                   'app, site or ext.'
    HELP_A       = 'Mandatory when target is [site] or [ext], specify here ' \
                   'the absolute path to the related app (if the app is to ' \
                   'be found locally), or a distant app specifier of the ' \
                   'form %s. Currently, a distant specifier must point to a ' \
                   'Subversion repository.' % APP_S
    HELP_P       = '[site] the port on which the site will listen. Defaults ' \
                   'to 8000.'
    HELP_O       = "[site, unix/linux only] the owner of the site. If " \
                   "specified, the whole site tree will be chown'ed to this " \
                   "owner. Optionally, you may specify the group by " \
                   "suffixing the owner by a colon and the group name, ie: " \
                   "appy:appy."
    HELP_D       = '[site] dependencies, as a series of local paths or ' \
                   'specifiers. For every local path, a symlink will be ' \
                   'created in <site>/lib. For every specifier, a local copy ' \
                   'will created in <site>/lib.'
    HELP_E       = '[site] The absolute path to the optional extension to ' \
                   'the app specified via option "-a" (if the ext is to be ' \
                   'found locally) or a distant ext specifier of the form ' \
                   '"%s".' % APP_S

    # Error messages
    WRONG_TARGET = 'Wrong target "%s".'
    F_EXISTS     = '%s already exists. You must specify an inexistent path ' \
                   'that this script will create and populate with a fresh %s.'
    NO_APP       = 'No app was specified.'
    WRONG_APP    = '%s does not exist or is not a folder.'
    WRONG_APP_S  = 'Wrong app specifier: %s'
    APP_S_KO     = 'An app specifier must be of the form "%s".' % APP_S
    SPEC_APP_KO  = "In that context, the app's local path must be given."

    def defineArguments(self):
        '''Define the allowed arguments for this program'''
        add = self.parser.add_argument
        # Positional arguments, common to all targets (app and site)
        add('target', help=Make.HELP_TARGET)
        add('folder', help=Make.HELP_FOLDER)
        # Optional arguments specific to "site"
        add('-a', '--app',  dest='app',  help=Make.HELP_A)
        add('-p', '--port', dest='port', help=Make.HELP_P)
        add('-o', '--owner', dest='owner', help=Make.HELP_O)
        add('-d', '--dependencies', dest='dependencies', nargs='+',
            help=Make.HELP_D)
        add('-e', '--ext',  dest='ext',  help=Make.HELP_E)

    def analyseApp(self, app, isExt=False, specifierDisallowed=False):
        '''Get the path to a local app (or ext if p_isExt is True) or the
           distant specifier else.'''
        if not app and not isExt: self.exit(self.NO_APP)
        if ',' in app:
            # A specifier
            if specifierDisallowed: self.exit(self.SPEC_APP_KO)
            r = parseSpecifier(app)
            if not r.url:
                print(self.WRONG_APP_S % app)
                self.exit(self.APP_S_KO % app)
        else:
            # It must be the path to a local folder
            r = Path(app).resolve()
            # This path must exist
            if not r.is_dir():
                self.exit(self.WRONG_APP % r)
        return r

    def analyseDependencies(self, dependencies):
        '''Analyses the specified p_dependencies and converts, in it, paths to
           Path instances and specifiers to Object instances.'''
        if not dependencies: return []
        i = 0
        length = len(dependencies)
        while i < length:
            dep = dependencies[i]
            if ',' in dep:
                # A specifier
                dep = parseSpecifier(dep)
                if not dep.url:
                    print(self.WRONG_APP_S % dep)
                    self.exit(self.APP_S_KO % dep)
            else:
                dep = Path(dep).resolve()
            dependencies[i] = dep
            i += 1
        return dependencies

    def analyseArguments(self):
        '''Check and store arguments'''
        # "folder" will be a Path created from the "folder" argument
        self.folder = None
        # Site/ext-specific attributes
        self.app = None # The result of parsing the '-a' option
        # Check arguments
        args = self.args
        # Check target
        target = args.target
        if target not in self.allowedTargets:
            self.exit(self.WRONG_TARGET % target)
        # Get a Path for the "folder" argument
        self.folder = Path(self.args.folder).resolve()
        # Check site-specific arguments and values
        if target == 'site':
            # The path to the site must not exist
            if self.folder.exists():
                self.exit(self.F_EXISTS % (self.folder, 'site'))
            # Manage the app and its ext
            self.app = self.analyseApp(self.args.app)
            ext = self.args.ext
            self.ext = None if not ext else self.analyseApp(ext, isExt=True)
            # The port must be an integer value
            port = self.args.port
            if port and not port.isdigit():
                self.exit(self.PORT_KO % port)
            self.port = int(port) if port else 8000
            # An owner may be specified
            self.owner = self.args.owner
            # Manage dependencies
            self.dependencies = self.analyseDependencies(self.args.dependencies)
        elif target == 'app':
            # No more check for the moment. The folder can exist or not.
            pass
        elif target == 'ext':
            # The path to the ext must not exist (it is not possible to update
            # an ext, only create it).
            if self.folder.exists():
                self.exit(self.F_EXISTS % (self.folder, 'ext'))
            # The path to the related app must be given
            self.app = self.analyseApp(self.args.app, specifierDisallowed=True)

    def run(self):
        target = self.args.target
        if target == 'site':
            Site(self.folder, self.app, self.port, self.owner,
                 self.dependencies, self.ext).create()
        elif target == 'app':
            action = 'update' if self.folder.exists() else 'create'
            eval('App(self.folder).%s()' % action)
        elif target == 'ext':
            Ext(self.folder, self.app).create()
        print('Done.')

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
if __name__ == '__main__': Make().run()
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# -*- coding: utf-8 -*-
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------
from appy.px import Px
from appy.fields import Field
from appy.gen.indexer import ValidValue

# ------------------------------------------------------------------------------
INVALID_MULTILINGUAL_VALUE = 'Multilingual field "%s" accepts a dict whose ' \
  'keys are in field.languages and whose values are strings.'

# ------------------------------------------------------------------------------
class Multilingual:
    '''Mixin class injected into any Field whose content can be multilingual'''

    # Default values to render when field values are empty
    emptyDefault = {'view': '-', 'edit': ''}
    # Default top space (in pixels) to apply in pxLanguage
    lgTop = {'view': 1, 'edit': 3}

    # Note that multilinguality is a dynamic feature: field values can be
    # unilingual or multilingual, depending on some condition on the container
    # object itself, or on some site-specific configuration.

    # PX displaying the language code and name besides the part of the
    # multilingual field storing content in this language.
    pxLanguage = Px('''
     <td style=":'padding-top:%dpx' % field.lgTop[layoutType]" width="25px">
      <span class="language help"
            title=":ztool.getLanguageName(lg)">:lg.upper()</span>
     </td>''')

    # PX acting as a substitute for the field pxView. This PX determines if the
    # field content is multilingual or not. If the field is unilingual, it
    # simply calls a PX named "px<layoutType>Uni" on the field. Else, it renders
    # such a PX for every supported language, calling "px<layoutType>Uni" for
    # every language, and assembling the result according to the "languages
    # layout". The "Uni-" PX receives the current language as variable "lg". If
    # the field is unilingual, the received "lg" variable is None.
    pxView = pxEdit = pxCell = Px('''
     <x var="languages=field.getAttribute(obj, 'languages');
             multilingual=len(languages) &gt; 1;
             pxUni=getattr(field, 'px%sUni' % layoutType.capitalize())">

      <!-- Display the uni-lingual version of the field -->
      <x if="not multilingual" var2="lg=None">:pxUni</x>

      <!-- Display the multi-lingual version of the field -->
      <x if="multilingual"
         var2="languageLayout=field.getLanguagesLayout(layoutType)">

       <!-- Horizontally-layouted multilingual field -->
       <table if="languageLayout == 'horizontal'" width="100%"
              class=":(layoutType == 'cell') and 'no' or ''"
              var="count=len(languages)">
        <tr valign="top"><x for="lg in languages"><x>:field.pxLanguage</x>
         <td width=":'%d%%' % int(100.0/count)"
             var="requestValue=requestValue[lg]|None;
                  value=value[lg]|field.emptyDefault[layoutType]">:pxUni</td>
        </x></tr></table>

       <!-- Vertically-layouted multilingual field -->
       <table if="languageLayout == 'vertical'"
              class=":(layoutType == 'cell') and 'no' or ''">
        <tr valign="top" height="20px" for="lg in languages">
         <x>:field.pxLanguage</x>
         <td var="requestValue=requestValue[lg]|None;
                  value=value[lg]|field.emptyDefault[layoutType]">:pxUni</td>
       </tr></table>
      </x>
     </x>''')

    def __init__(self, languages, languagesLayouts):
        '''Inject multilingual-specific attributes on p_self'''
        # If "languages" holds more than one language, the field will be
        # multi-lingual and several widgets will allow to edit/visualize the
        # field content in all the supported languages. The field is also used
        # by the CK spell checker.
        self.languages = languages
        # When content exists in several languages, how to render them? Either
        # horizontally (one below the other), or vertically (one besides the
        # other). Specify here a dict whose keys are layouts ("edit", "view")
        # and whose values are either "horizontal" or "vertical".
        self.languagesLayouts = languagesLayouts

    def isMultilingual(self, obj, dontKnow=False):
        '''Is this field multilingual ? If we don't know, say p_dontKnow'''
        if not obj:
            if callable(self.languages):
                # In that case, it is impossible to know
                return dontKnow
            else: return len(self.languages) > 1
        return len(self.getAttribute(obj, 'languages')) > 1

    def getLanguagesLayout(self, layoutType):
        '''Gets the way to render a multilingual field on p_layoutType.'''
        if self.languagesLayouts and (layoutType in self.languagesLayouts):
            return self.languagesLayouts[layoutType]
        # Else, return a default value that depends of the format
        return self.defaultLanguagesLayouts[layoutType]

    def getCopyValue(self, obj):
        '''A value being multilingual is stored in a dict. For such a value,
           standard method Field.getCopyValue must return a distinct copy of the
           value as stored on p_obj.'''
        r = self.getValue(obj)
        if isinstance(r, dict): r = r.copy()
        return r

    def valueIsInRequest(self, obj, request, name=None, layoutType='view'):
        '''Multilingual values are stored in specific input fields with
           specific names.'''
        # If we are on the search layout, p_obj, if not None, is certainly not
        # the p_obj we want here (can be a home object).
        if layoutType == 'search':
            return Field.valueIsInRequest(self, obj, request, name, layoutType)
        languages = self.getAttribute(obj, 'languages')
        if len(languages) == 1:
            return Field.valueIsInRequest(self, obj, request, name, layoutType)
        # Is is sufficient to check that at least one of the language-specific
        # values is in the request.
        return request.has_key('%s_%s' % (name, languages[0]))

    def getRequestValue(self, obj, requestName=None):
        '''The request value may be multilingual'''
        request = obj.REQUEST
        name = requestName or self.name
        languages = self.getAttribute(obj, 'languages')
        # A unilingual field
        if len(languages) == 1: return request.get(name, None)
        # A multilingual field
        r = {}
        for language in languages:
            r[language] = request.get('%s_%s' % (name, language), None)
        return r

    def isEmptyValue(self, obj, value):
        '''Returns True if the p_value must be considered as an empty value'''
        if not isinstance(value, dict):
            return Field.isEmptyValue(self, obj, value)
        # p_value is a dict of multilingual values. For such values, as soon
        # as a value is not empty for a given language, the whole value is
        # considered as not being empty.
        for v in value.itervalues():
            if not Field.isEmptyValue(self, obj, v): return
        return True

    def isCompleteValue(self, obj, value):
        '''Returns True if the p_value must be considered as complete. For a
           unilingual field, being complete simply means not being empty. For a
           multilingual field, being complete means that a value is present for
           every language.'''
        if not self.isMultilingual(obj):
            return Field.isCompleteValue(self, obj, value)
        # As soon as a given language value is empty, the global value is not
        # complete.
        if not value: return True
        for v in value.itervalues():
            if Field.isEmptyValue(self, obj, v): return
        return True

    def getFormattedValue(self, obj, value, layoutType='view',
                          showChanges=False, language=None):
        '''The multilingual and unilingual variants of p_value's formatted
           version differ.'''
        # Note that p_language represents the UI language, while variable
        # "languages" below represents the content language(s) of this field.
        languages = self.getAttribute(obj, 'languages')
        uni = self.getUniFormattedValue
        if (len(languages) == 1) or isinstance(value, basestring):
            # Normally, p_value should not be a string if there is a single
            # language. This can happen in exceptional cases, ie, in a
            # object's history (data change), when an object was transmitted
            # from one App1 to App2, where a field is unilingual in App1 and
            # multilingual in App2.
            return uni(obj, value, layoutType,
                       showChanges, userLanguage=language)
        # Return the dict of values whose individual, language-specific values
        # have been formatted via m_getUniFormattedValue.
        if not value and not showChanges: return value
        r = {}
        for lg in languages:
            if not value: val = ''
            else: val = value[lg]
            r[lg] = uni(obj, val, layoutType, showChanges, language=lg)
        return r

    def getShownValue(self, obj, value, layoutType='view',
                      showChanges=False, language=None):
        '''For a multilingual field, this method only shows one specific
           language part.'''
        # Be careful: p_language represents the UI language, while variable
        # "languages" below represents the content language(s) of this field.
        languages = self.getAttribute(obj, 'languages')
        uni = self.getUniFormattedValue
        if len(languages) == 1:
            return uni(obj, value, layoutType,
                       showChanges, userLanguage=language)
        if not value: return value
        # Try to propose the part that is in the user language, or the part of
        # the first content language else.
        lg = obj.getUserLanguage()
        if lg not in value: lg = languages[0]
        return uni(obj, value[lg], layoutType, showChanges, language=lg)

    def getIndexValue(self, obj, forSearch=False):
        '''Multilingual content must be concatenated into a single value to be
           indexed.'''
        # Must we produce an index value?
        if not self.getAttribute(obj, 'mustIndex'):
            return ValidValue.forString(None, forSearch)
        # Get the field value on p_obj
        r = self.getValue(obj)
        # Possibly transform the value
        if self.indexValue: r = self.indexValue(obj.appy(), r)
        # Manage multilinguality
        if isinstance(r, dict):
            r = ' '.join([self.getUniIndexValue(obj, v, forSearch) \
                          for v in r.itervalues()])
        else:
            r = self.getUniIndexValue(obj, r, forSearch)
        return ValidValue.forString(r, forSearch)

    def getStorableValue(self, obj, value, complete=False):
        languages = self.getAttribute(obj, 'languages')
        if len(languages) == 1:
            return self.getUniStorableValue(obj, value)
        # A multilingual value is stored as a dict whose keys are ISO 2-letters
        # language codes and whose values are strings storing content in the
        # language ~{s_language: s_content}~.
        if not value: return
        for lg in languages:
            value[lg] = self.getUniStorableValue(obj, value[lg])
        return value

    def store(self, obj, value):
        '''Stores p_value on p_obj for this field'''
        languages = self.getAttribute(obj, 'languages')
        if (len(languages) > 1) and value and \
           (not isinstance(value, dict) or (len(value) != len(languages))):
            raise Exception(INVALID_MULTILINGUAL_VALUE % self.name)
        Field.store(self, obj, value)

    def storeValueFromAjax(self, obj, value, previousData):
        '''Stores p_value on p_obj, taking care of multilinguality'''
        # Call the base method if p_self is not multilingual
        if not self.isMultilingual(obj):
            return Field.storeValueFromAjax(self, obj, value, previousData)
        # Manage a multilingual field
        if previousData is not None:
            # Take a copy of previousData because it is mutable (dict)
            prevData = previousData[self.name]
            if prevData is not None: prevData = prevData.copy()
            previousData[self.name] = prevData
        # We get a partial value, for one language only
        language = obj.REQUEST['languageOnly']
        v = self.getUniStorableValue(obj, value)
        getattr(obj.aq_base, self.name)[language] = v
        return ' (%s)' % language
# ------------------------------------------------------------------------------



'''Validation-related stuff'''

# ------------------------------------------------------------------------------
from appy.px import Px
from appy import Object as O
from appy.shared import utils as sutils

# ------------------------------------------------------------------------------
class Validator:
    '''Class responsible for validating data encoded by the user in the UI'''
    FIELD_NOT_FOUND = 'field %s::%s was not found.'

    # PX rendering the "yellow box" containing validation errors
    pxErrors = Px('''
     <table class="small" width="100%">
      <tr><th>:_('field_name')</th><th>:_('error_name')</th></tr>
      <tr for="labels, message in validator.getGroupedErrors()">
       <td>::labels</td><td>:message</td></tr>
     </table>''')

    def __init__(self, obj):
        # A validator is created every time an p_obj(ect) is created or updated
        # from the UI via a web form.
        self.obj = obj
        # The (sub/standard/switch) fields whose widgets have received data from
        # the UI.
        self.fields = []
        self.fieldsByName = {}
        # The field values extracted from the request and converted
        self.values = O() # ~{O(s_fieldName=fieldValue)}~
        # When validation errors are encountered, they are stored in the
        # following dict. For example, if an error occurs on field named
        # "field1", an entry at key "field1" will be stored on this instance,
        # holding the translated error message.
        self.errors = O() # ~{O(s_fieldName=s_errorMessage)}~

    def intraFieldValidation(self, fields=None):
        '''This method performs field-specific validation for every field from
           the page that is being created or edited. For every field whose
           validation generates an error, we add an entry in p_self.errors. For
           every field, we add in p_self.values an entry with the
           "ready-to-store" field value.

           Il p_fields are given, we use them instead of getting currently
           edited fields on the current page for p_self.obj. ie, p_fields can be
           fields from a switch.
        '''
        obj = self.obj
        rq = obj.REQUEST
        page = rq.form.get('page')
        for field in obj.getAppyTypes('edit', page, fields=fields):
            if not field.validable or not field.isClientVisible(obj): continue
            value = field.getRequestValue(obj)
            # Perform individual field validation
            message = field.validate(obj, value)
            if message:
                setattr(self.errors, field.name, message)
            else:
                setattr(self.values, field.name,
                        field.getStorableValue(obj, value))
            # Store this field
            self.fields.append(field)
            self.fieldsByName[field.name] = field
            # Validate inner fields within outer fields
            if field.outer: field.subValidate(obj, value, self.errors)
            # Validate fields inside switch fields
            if field.type == 'Switch':
                fieldset, subFields = field.getChosenFields(obj, fieldset=value)
                if subFields:
                    self.intraFieldValidation(fields=subFields)

    def interFieldValidation(self):
        '''This method is called when individual validation of all fields
           succeeds (when editing or creating an object, see
           m_intraFieldValidation). Then, this method performs inter-field
           validation. This way, the user must first correct individual fields
           before being confronted to potential inter-field validation
           errors.'''
        obj = self.obj.appy()
        if not hasattr(obj, 'validate'): return
        msg = obj.validate(self.values, self.errors)
        # Those custom validation methods may have added fields in the given
        # p_errors object. Within this object, for every error message that is
        # not a string, we replace it with the standard validation error for the
        # corresponding field.
        for k, v in self.errors.d().iteritems():
            if not isinstance(v, basestring):
                setattr(self.errors, k, obj.translate('field_invalid'))
        return msg

    def addMessage(self, r, messages, label, message):
        '''Used by m_getErrors, this method adds, in p_r, the error p_message
           about a field whose label is p_label. Indexes of messages already
           stored in p_r are in dict p_messages.'''
        if message in messages:
            # We have already encountered this message
            r[messages[message]][0] += '<br/>%s' % label
        else:
            r.append([label, message])
            messages[message] = len(r) - 1

    def getGroupedErrors(self):
        '''Returns the list of all errors, grouped by message: if the same error
           message concerns several fields, a single entry is inserted in the
           list.'''
        _ = self.obj.o.translate
        obj = self.obj.appy()
        # A list of tuples (labels, message): "labels" being the list of field
        # labels whose error is described in "message".
        r = []
        # Remember, for every message stored in p_r, its index
        messages = {} # ~{s_message: i_index}~
        # Walk fields in the order of their definition, via self.fields
        for field in self.fields:
            # Is there an error for this field ?
            message = self.errors.get(field.name)
            if not message: continue
            label = _('label', field=field)
            self.addMessage(r, messages, label, message)
        # Add summarized messages for compound fields
        compound = []
        summaryMessage = _('validation_sub_error')
        for name, message in self.errors.d().iteritems():
            # Ignore already-managed, not-compound fields
            if '*' not in name: continue
            name = name.split('*', 1)[0]
            # Ignore already-encountered compound (sub-)fields
            if name in compound: continue
            compound.append(name)
            # Get the label for this compound field
            field = self.fieldsByName.get(name)
            if not field:
                label = name
                obj.log(Validator.FIELD_NOT_FOUND % (obj.klass.__name__, name),
                        type='error')
            else:
                label = _('label', field=field)
            self.addMessage(r, messages, label, summaryMessage)
        return r

    def getJsErrors(self):
        '''Generates the definition of a JavaScript associative array
           containing all errors collected in p_self.errors.'''
        return self.errors and sutils.getStringFrom(self.errors.d()) or 'null'
# ------------------------------------------------------------------------------



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.px import Px
from appy import ui, utils

# Error messages - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
TB_COLS_ERR = 'A tabs-style group must have a single column.'
G_COLS_ERR  = 'For a grid-style group, you must specify an even number of ' \
              'columns, in order to produce couples of label/field columns. ' \
              'For every couple, the first columns holds a field label and ' \
              'the second one holds the field content.'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Group:
    '''Used for describing a group of fields within a page'''
    def __init__(self, name, columns=None, wide=True, style='section2',
      hasLabel=True, hasDescr=False, hasHelp=False, hasHeaders=False,
      group=None, colspan=1, align='center', valign='top', css_class='',
      labelCss=None, master=None, masterValue=None, cellpadding=1,
      cellspacing=1, cellgap='0.6em', label=None, translated=None):
        self.name = name
        # In its simpler form, field "columns" below can hold a list or tuple
        # of column widths expressed as strings, that will be given as is in
        # the "width" attributes of the corresponding "td" tags. Instead of
        # strings, within this list or tuple, you may give Column instances
        # (see below).
        self.columns = columns
        self._setColumns(style)
        # If field "wide" below is True, the HTML table corresponding to this
        # group will have width 100%. You can also specify some string value,
        # which will be used for HTML param "width".
        if wide == True:
            self.wide = '100%'
        elif isinstance(wide, str):
            self.wide = wide
        else:
            self.wide = ''
        # Groups of various styles can be rendered. If "style" is:
        # - 'sectionX'  (X can be 1, 2, 3...) 
        #               the group will be rendered as a "section": the group
        #               title will be rendered in some style (depending on X)
        #               before the widgets;
        # - 'fieldset'  all widgets within the group will be rendered within an
        #               HTML fieldset;
        # - 'tabs'      the group will be rendered as tabs. One tab will be
        #               rendered for every inner widget. If you want some tab to
        #               contain several widgets, specify a group as sub-group of
        #               the group having style 'tabs';
        # - 'grid'      the widgets will be rendered in some standardized,
        #               tabular way.
        self.style = style
        # If hasLabel is True, the group will have a name and the corresponding
        # i18n label will be generated.
        self.hasLabel = hasLabel
        # If hasDescr is True, the group will have a description and the
        # corresponding i18n label will be generated.
        self.hasDescr = hasDescr
        # If hasHelp is True, the group will have a help text associated and the
        # corresponding i18n label will be generated.
        self.hasHelp = hasHelp
        # If hasheaders is True, group content will begin with a row of headers,
        # and a i18n label will be generated for every header.
        self.hasHeaders = hasHeaders
        self.nbOfHeaders = len(self.columns)
        # If this group is himself contained in another group, the following
        # attribute is filled.
        self.group = Group.get(group)
        # If the group is rendered into another group, we can specify the number
        # of columns that this group will span.
        self.colspan = colspan
        self.align = align
        self.valign = valign
        self.cellpadding = cellpadding
        self.cellspacing = cellspacing
        # Beyond standard cellpadding and cellspacing, cellgap can define an
        # additional horizontal gap between cells in a row. So this value does
        # not add space before the first cell or after the last one.
        self.cellgap = cellgap
        if style == 'tabs':
            # Group content will be rendered as tabs. In this case, some
            # param combinations have no sense.
            self.hasLabel = self.hasDescr = self.hasHelp = False
            # Inner field/group labels will be used as tab labels
        # "css_class", if specified, will be applied to the whole group
        self.css_class = css_class
        # "labelCss" is the CSS class that will be applied to the group label
        self._setLabelCss(labelCss, style)
        self.master = master
        self.masterValue = utils.initMasterValue(masterValue)
        if master: master.slaves.append(self)
        self.label = label # See similar attr of Type class
        # If a translated name is already given here, we will use it instead of
        # trying to translate the group label.
        self.translated = translated

    def __repr__(self):
        '''p_self's string representation'''
        parent = ',parent=%s' % self.group.name if self.group else ''
        return '<group %s%s>' % (self.name, parent)

    def _setColumns(self, style):
        '''Standardizes field "columns" as a list of Column instances. Indeed,
           the initial value for field "columns" may be a list or tuple of
           Column instances or strings.'''
        # Start with a default value, if self.columns is None
        if not self.columns:
            if style == 'grid':
                # One column for the labels, another for the remaining elements
                self.columns = ['150em', '']
            else:
                self.columns = ['100%']
        # Standardize columns as a list of Column instances
        for i in range(len(self.columns)):
            columnData = self.columns[i]
            if not isinstance(columnData, Column):
                self.columns[i] = Column(self.columns[i])
        # Standardize or check columns depending on group style
        if style == 'tabs':
            # There must be a unique column
            if len(self.columns) > 1: raise Exception(TB_COLS_ERR)
        elif style == 'grid':   
            # grid has always an even number of columns (couples of label/field
            # columns)
            if len(self.columns) % 2: raise Exception(G_COLS_ERR)

    def _setLabelCss(self, labelCss, style):
        '''For "sectionX"-style groups, the applied CSS class is by default the
           p_style itself. For the other styles of groups (tabs, grids,...) it
           is not the case.'''
        if labelCss:
            self.labelCss = labelCss
        else:
            # Set a default value for the label CSS
            if style.startswith('section'):
                self.labelCss = style
            else:
                self.labelCss = 'section3'

    @staticmethod
    def get(groupData):
        '''Produces a Group instance from p_groupData. User-defined p_groupData
           can be a string or a Group instance; this method returns always a
           Group instance.'''
        res = groupData
        if res and isinstance(res, str):
            # Group data is given as a string. 2 more possibilities:
            # (a) groupData is simply the name of the group;
            # (b) groupData is of the form <groupName>_<numberOfColumns>.
            groupElems = groupData.rsplit('_', 1)
            if len(groupElems) == 1:
                res = Group(groupElems[0])
            else:
                try:
                    nbOfColumns = int(groupElems[1])
                except ValueError:
                    nbOfColumns = 1
                width = 100.0 / nbOfColumns
                res = Group(groupElems[0], ['%.2f%%' % width] * nbOfColumns)
        return res

    def getMasterData(self):
        '''Gets the master of this group (and masterValue) or, recursively, of
           containing groups when relevant.'''
        if self.master: return self.master, self.masterValue
        if self.group: return self.group.getMasterData()

    def insertInto(self, r, groups, page, className, content='fields'):
        '''Insert the UiGroup instance corresponding to this Group instance
           into p_r, the recursive structure used for displaying elements in a
           given p_page (fields, searches, transitions...) and return this
           UiGroup instance.'''
        # First, create the corresponding UiGroup if not already in p_groups
        if self.name not in groups:
            uiGroup= groups[self.name] = UiGroup(self, page, className, content)
            # Insert the group at the higher level (ie, directly in p_r) if the
            # group is not itself in a group.
            if not self.group:
                r.append(uiGroup)
            else:
                outerGroup = self.group.insertInto(r, groups, page,
                                                   className, content)
                outerGroup.addElement(uiGroup)
        else:
            uiGroup = groups[self.name]
        return uiGroup

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Column:
    '''Used for describing a column within a Group like defined above'''
    def __init__(self, width, align="left"):
        self.width = width
        self.align = align

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class UiGroup:
    '''On-the-fly-generated data structure that groups all elements
       (fields, searches, transitions...) sharing the same Group instance, that
       the currently logged user can see.'''

    # Render a help icon for a group
    pxHelp = Px(lambda c: '<abbr title="%s"><img class="iconS" src="%s"/>' \
                    '</abbr>' % (c._('help', field=c.field), c.url('help.svg')))

    # Render the group title, description and help
    pxHeader = Px('''
     <!-- Title -->
     <tr if="field.hasLabel">
      <td colspan=":len(field.columnsWidths)" class=":field.labelCss"
          align=":dleft">
       <x>::_(field.labelId)</x><x if="field.hasHelp">:field.pxHelp</x>
      </td>
     </tr>
     <tr if="field.hasDescr">
      <td colspan=":len(field.columnsWidths)"
          class="discreet">::_(field.descrId)</td>
     </tr>''')

    # Render the fields within a group in the most frequent cases:
    # style = "sectionX" or "fieldset". The group is referred as var "field".
    pxFields = Px('''
     <table var="cellgap=field.cellgap" width=":field.wide"
            align=":ui.Language.flip(field.align, dir)"
            id=":tagId" name=":tagName" class=":groupCss"
            cellspacing=":field.cellspacing" cellpadding=":field.cellpadding">
      <!-- Title, description and help -->
      <x if="field.showHeader()">:field.pxHeader</x>
      <!-- The column headers -->
      <tr>
       <th for="colNb in range(len(field.columnsWidths))"
           align=":ui.Language.flip(field.columnsAligns[colNb], dir)"
           width=":field.columnsWidths[colNb]">::field.hasHeaders and \
            _('%s_col%d' % (field.labelId, (colNb+1))) or ''</th>
      </tr>
      <!-- The rows of widgets -->
      <tr valign=":field.valign" for="row in field.elements">
       <td for="field in row" colspan=":field.colspan|1"
           style=":not loop.field.last and ('padding-right:%s'% cellgap) or ''">
        <x if="field">:field.pxRender</x>
       </td>
      </tr>
     </table>''')

    # Render a group with style = 'fieldset'
    pxFieldset = Px('''
     <fieldset>
      <legend if="field.hasLabel">
       <i>::_(field.labelId)></i><x if="field.hasHelp">:field.pxHelp</x>
      </legend>
      <div if="field.hasDescr" class="discreet">::_(field.descrId)</div>
      <x>:field.pxFields</x>
     </fieldset>''')

    # Render a group with style = 'tabs'
    pxTabs = Px('''
     <table width=":field.wide" class=":groupCss" id=":tagId" name=":tagName">
      <!-- First row: the tabs -->
      <tr valign="middle"><td class="tabSep tabBottom">
       <table class="tabs" id=":'tabs_%s' % field.name">
        <tr valign="middle">
         <x for="sub in field.elements"
            var2="suffix='%s_%s' % (field.name, sub.name);
                  tabId='tab_%s' % suffix">
          <td class="tab" id=":tabId">
           <a onclick=":'showTab(%s)' % q(suffix)"
              class="clickable">:_(sub.labelId)</a>
          </td>
         </x>
        </tr>
       </table>
      </td></tr>

      <!-- Other rows: the fields -->
      <tr for="sub in field.elements"
          id=":'tabcontent_%s_%s' % (field.name, sub.name)"
          style=":(loop.sub.nb==0) and 'display:table-row' or 'display:none'">
       <td var="field=sub">:field.pxRender</td>
      </tr>
     </table>
     <script>:'initTab(%s,%s,%s)' % \
      (q('tab_%s'%field.name), q('%s_%s'%(field.name, field.elements[0].name)),\
       'true' if o.isTemp() else 'false')</script>''')

    # Render a group with style = 'grid'
    pxGrid = Px('''
     <table cellpadding="0" cellspacing="0" width=":field.wide"
            class=":groupCss" id=":tagId" name=":tagName" align=":field.align">
      <!-- Title, description and help -->
      <x if="field.showHeader()">:field.pxHeader</x>
      <tr><th for="col in field.columns" width=":col.width"></th></tr>
      <tr for="row in field.elements" valign="top"
          class=":loop.row.odd and 'odd' or 'even'">
       <x for="sub in row">
        <td id="summaryCell">
         <label if="sub.hasLabel and \
                    (sub.type != 'Action')">::_('label', field=sub)</label></td>
        <td var="field=sub" id="summaryCell">:field.pxRender</td>
       </x>
       <!-- Complete the last row when relevant -->
       <x if="loop.row.last"
          for="i in range(int((len(field.columns)/2))-len(row))">
         <td></td><td></td>
       </x>
      </tr>
     </table>''')

    # PX that renders a group of fields (the group is referred as var "field")
    pxRender = Px('''
     <x var="tagCss=field.master and ('slave*%s*%s' % \
                    (field.master.getMasterTag(layout), \
                    '*'.join(field.masterValue))) or '';
             widgetCss=field.css_class;
             groupCss=tagCss and ('%s %s' % (tagCss, widgetCss)) or widgetCss;
             tagName=field.master and 'slave' or '';
             tagId='%d_%s' % (o.iid, field.name)">:field.pxFromStyle()</x>''')

    # PX that renders a group of searches
    pxViewSearches = Px('''
     <x var="collapse=field.getCollapseInfo(field.labelId, req)">
      <!-- Group name, prefixed by the expand/collapse icon -->
      <div class="portletGroup"><x>:collapse.px</x>
       <x>:field.translated or _(field.labelId)</x>
      </div>
      <!-- Group content -->
      <div id=":collapse.id" style=":'padding-left: 10px; %s' % collapse.style">
       <x for="searches in field.elements">
        <x for="elem in searches">
         <!-- An inner group within this group -->
         <x if="elem.type== 'group'" var2="field=elem">:field.pxViewSearches</x>
         <!-- A search -->
         <x if="elem.type != 'group'" var2="search=elem">:search.view</x>
        </x>
       </x>
      </div></x>''')

    # PX that renders a group of transitions
    pxViewTransitions = Px('''
     <!-- Render a group of transitions, as a one-column table -->
     <table>
      <x for="row in trGroup.elements">
       <x for="transition in row"><tr><td>:transition.px</td></tr></x>
      </x>
     </table>''')

    # What PX to use, depending on group content?
    pxByContent = {'fields': pxRender, 'searches': pxViewSearches,
                   'transitions': pxViewTransitions}

    def __init__(self, group, page, className, content='fields'):
        '''A UiGroup can group various kinds of elements: fields, searches,
           transitions... The type of content that one may find in this group
           is given in p_content.
           * p_group      is the Group instance corresponding to this UiGroup;
           * p_page       is the Page instance where the group is rendered (for
                          transitions, it corresponds to a virtual page
                          "workflow");
           * p_className  is the name of the class that holds the elements to
                          group.'''
        self.type = 'group'
        # All p_group attributes become self attributes. This is required
        # because a UiGroup, in some PXs, must behave like a Field (ie, have
        # the same attributes, like "master".
        for name, value in group.__dict__.items():
            if not name.startswith('_'):
                setattr(self, name, value)
        self.group = group
        self.columnsWidths = [col.width for col in group.columns]
        self.columnsAligns = [col.align for col in group.columns]
        # The name of the page where the group lies
        self.page = page.name
        # The elements (fields or sub-groups) contained in the group, that the
        # current user may see. They will be inserted by m_addElement below.
        self.flatElements = self.style == 'tabs'
        if self.flatElements:
            # Elements will be stored as a simple list
            self.elements = []
        else:
            # In most cases, "elements" will be a list of lists for rendering
            # them as a table.
            self.elements = [[]]
        # PX to use for rendering this group
        self.px = self.pxByContent[content]
        # Names of i18n labels for this group
        if not self.hasLabel and not self.hasDescr and not self.hasHelp: return
        labelName = self.name
        prefix = None
        if group.label:
            if isinstance(group.label, str): prefix = group.label
            else: # It is a tuple (className, name)
                if group.label[1]: labelName = group.label[1]
                if group.label[0]: prefix = group.label[0]
        if not prefix:
            prefix = '%s_group' % className
        self.labelId = '%s_%s' % (prefix, labelName)
        self.descrId = self.labelId + '_descr'
        self.helpId  = self.labelId + '_help'

    def addElement(self, element):
        '''Adds p_element into self.elements. We try first to add p_element into
           the last row. If it is not possible, we create a new row.'''
        if self.flatElements:
            self.elements.append(element)
            return
        # Get the last row
        lastRow = self.elements[-1]
        numberOfColumns = len(self.columnsWidths)
        # Grid groups span a single field on 2 columns
        if self.style == 'grid': numberOfColumns = numberOfColumns / 2
        # Compute the number of columns already filled in the last row
        filledColumns = 0
        for rowElem in lastRow: filledColumns += rowElem.colspan
        freeColumns = numberOfColumns - filledColumns
        if freeColumns >= element.colspan:
            # We can add the element in the last row
            lastRow.append(element)
        else:
            if freeColumns:
                # Terminate the current row by appending empty cells
                for i in range(freeColumns): lastRow.append('')
            # Create a new row
            self.elements.append([element])

    def getCollapseInfo(self, id, request):
        '''Returns a Collapsible instance, that determines if this group,
           represented as an expandable menu item, is collapsed or expanded.'''
        return ui.Collapsible(id, request)

    def pxFromStyle(self):
        '''Get the PX to use for rendering a group, depending on its style'''
        style = self.style
        if style[-1].isdigit(): return self.pxFields # sectionX
        px = 'px%s' % style.capitalize()
        return getattr(self, px)

    def showHeader(self):
        '''The block "title, description, help" must not be rendered even if it
           exists, because it is rendered elsewhere.'''
        if self.style == 'fieldset': return
        parent = self.group.group
        if parent and (parent.style in ('tabs', 'grid')): return
        return self.hasLabel or self.hasDescr
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.px import Px
from appy.utils import noSwitchLayouts

from appy.server.cookie import Cookie

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class AuthenticationContext:
    '''When an application uses an authentication context, its users, when
       logging in, must, besides their login and password, choose a context
       among possible authentication contexts. Then, your application has access
       to the chosen context via the handler attribute "authContext".

       If you want to use authentication contexts in your Appy application, you
       must create a class that inherits from this one and overrides some of its
       methods (see below). Then, in your Config instance set an instance of
       your class in attribute config.security.authContext.
    '''
    traverse = {}

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    #                           W a r n i n g
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    # Several methods defined hereafter can be overridden by child classes
    # written by developers in Appy apps. These methods are called from within
    # the Appy authentication mechanism:
    #
    # getContexts
    # getDefaultContext
    # getObject

    # It means that, when these methods are called, security-related elements
    # may not be available yet. For example, within these overridden methods, do
    # not use property "user" on the tool or any other object: it has not been
    # set yet. Use, wherever available, parameter "user" instead. Do not use
    # anything requiring security mechanisms: they may not be ready yet (ie,
    # calls to methods like "do" or "search" with secure=True as parameter).

    # Methods starting with an underscore should not be overridden.

    # Part of the login form for choosing the context
    pxOnLogin = Px('''
     <x>:_('login_context')</x>
     <select id="__ac_ctx" name="__ac_ctx"
             var="ctxDefault=ctx.getDefaultContext(tool)">
      <option value="">:_('choose_a_value')</option>
      <option for="opt in ctx.getContexts(tool)" value=":opt[0]"
              selected=":opt[0] == ctxDefault">:opt[1]</option>
     </select>''')

    # Zone where to display (or switch) the context when the user is logged
    pxLogged = Px('''
     <x var="showSwitchOptions=ctx._showSwitchOptions(tool, layout)">
      <x if="showSwitchOptions"
         var2="options,present=ctx._getSwitchOptions(tool);
               selected=guard.authContext or ''">
       <select if="options" class="discreet" onchange="switchContext(this)">
        <!-- Set this option only to be able to select a single option -->
        <option if="not present and (len(options) == 1)"></option>
        <option for="opt in options" value=":opt[0]"
                selected=":opt[0] == selected">:opt[1]</option>
       </select>
      </x>
      <!-- Display the unswitchable authentication context -->
      <x if="not showSwitchOptions and guard.authContext">
       <i>:guard.authContextName</i> 
      </x>
     </x>''',

     js='''
      function switchContext(select) {
        goto(siteUrl + '/tool/AuthContext/switch?ctx=' + select.value) }''')

    def __init__(self, chooseOnLogin=True, noContextPosition='start',
                 homeAfterSwitch=False):
        # Must the user choose its context when logging in ?
        self.chooseOnLogin = chooseOnLogin
        # When choosing a context is not mandatory (see m_isMandatory) and the
        # special entry "no context" can thus be selected, this attribute
        # indicates it must be inserted in the list of available contexts: at
        # the "start" or at the "end". If you want this special entry not to be
        # available at all, set it to None. Indeed, a context may be declared as
        # not mandatory because not selectable when logging in, but, once
        # logged, it may be required to prevent the user from selecting entry
        # "no context".
        self.noContextPosition = noContextPosition
        # Note that chooseOnLogin=False & switchContext=False (see below) is
        # useless.
        # ~~~
        # By default, after switching the context, the user stays on the same
        # page. If you want him to be redirected to its home page, set the
        # following attribute to True.
        self.homeAfterSwitch = homeAfterSwitch

    def _showSwitchOptions(self, tool, layout):
        '''In some situations, switch options are not to be shown'''
        # On some pages, selectors can't be shown
        if layout in noSwitchLayouts: return
        # No switch option if switching is not allowed
        if not self.switchContext(tool): return
        return True

    def _getSwitchOptions(self, tool):
        '''Returns a tuple (options, present), where:
           - "options" is the list of values for the "switch" widget (=the list
             of possible contexts for the user);
           - "present" is a boolean value indicating if the current
             authentication context is in this list.
        '''
        # Get the available contexts
        r = self.getContexts(tool, tool.user) or []
        # Add "no context" switch option when relevant
        if not self.isMandatory(tool) and self.noContextPosition:
            no = ('', tool.translate('everything'))
            if self.noContextPosition == 'end':
                r.append(no)
            else: # 'start'
                r.insert(0, no)
        # Is the current authentication context in this list ?
        ctx = tool.guard.authContext or ''
        if not ctx and not r:
            # We consider the context to be present
            present = True
        else:
            present = False
            for id, name in r:
                if id == ctx:
                    present = True
                    break
        return r, present

    def _maySwitchTo(self, tool, option):
        '''Is p_option a valid option to switch to ?'''
        options, present = self._getSwitchOptions(tool)
        if not options: return
        for id, title in options:
            if id == option: return True

    def _check(self, tool):
        '''Returns True if there is no problem with the current configuration
           related to the authentication context.'''
        # Check that the current context is among the valid switch options
        options, present = self._getSwitchOptions(tool)
        return present

    def _setOnGuard(self, tool, user, guard, ctx):
        '''Set, on the the p_guard, authentication-p_ctx-related attributes'''
        # The authentication context in itself (a short identifier)
        guard.authContext = ctx
        # Its human-readable name
        guard.authContextName = self.getName(tool, user, ctx)
        # The corresponding object, if any
        guard.authObject = self.getObject(tool, user, ctx)

    def isMandatory(self, tool):
        '''When authentication contexts are activated, is the user forced to
           choose one ?'''
        return True

    def switchContext(self, tool):
        '''Is the user allowed to switch context once logged ?'''
        return True

    def getContexts(self, tool, user):
        '''Returns the application-specific authentication contexts, as a list
           of tuples (s_context, s_name). s_context is a short string that
           identifies the context, while s_name is a human-readable name that
           will be shown in the UI.'''

    def getDefaultContext(self, tool, user):
        '''Returns the default context among contexts as returned by
           m_getContexts.'''
        # Does not need to be overridden if there is no default context

    # This method must not be overridden
    def getName(self, tool, user, context):
        '''Returns the name of some given p_context'''
        for ctx, name in self.getContexts(tool, user):
            if ctx == context:
                return name

    def getObject(self, tool, user, context):
        '''If an authentication context corresponds to an object, this method
           must be overridden to return the object corresponding to the given
           p_context. Then, the object is available as attribute "authObject"
           on the guard.'''
        # Overriding this method optimizes performance. Indeed, your object is
        # "loaded" once per request and is available in guard.authObject: you
        # don't need to call o.geObject(context) every time you need to get
        # it.

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    #                           Class methods
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    traverse['switch'] = 'Authenticated'

    @classmethod
    def switch(class_, tool):
        '''Switches from one authentication context to another'''
        # Check that context switching is allowed
        authContext = tool.config.security.authContext
        req = tool.req
        ctx = req.ctx
        if not authContext or not authContext._maySwitchTo(tool, ctx):
            self.raiseUnauthorized()
        # Perform the switch by updating the authentication cookie and
        # user-related data structures stored on the guard.
        Cookie.update(tool.H(), ctx)
        authContext._setOnGuard(tool, tool.user, tool.guard, ctx)
        # Return to the URL specified in the request, or, by default:
        # a) return to the home page if we were displaying an error message or
        #    if required by the auhentication context;
        # b) else, stay on the same page (=go to the referer page).
        back = req.url
        if not back:
            back = tool.getHomePage() if authContext.homeAfterSwitch \
                                      else tool.referer
        tool.goto(back)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# -*- coding: utf-8 -*-
xhtmlInput = '''
<ul>
<li>Aa</li>
<li>Bb
 <ul>
  <li>Sub</li>
 </ul>
</li>
</ul>
<ol>
 <li>UN</li>
 <li>Deux<ol><li>SubDeux</li></ol></li>
 <li>Trois<ol><li>SubTrois</li></ol></li>
 <li>Quatre</li>
</ol>
<ol>
 <li>Numbered</li>
 <ol><li>Sub</li></ol>
 <li>Numbered<ol><li>Sub2</li></ol></li>
 <li>Numbered</li>
</ol>
<p>-</p>
<ol>
  <li>
    <ol>
      <li>
        <ol>
          <li><em><u>Contrôles des revenus</u></em></li>
        </ol>
      </li>
    </ol>
  </li>
</ol>
'''



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.px import Px
from appy.model.fields import Field
from appy.ui.layout import Layouts, Layout

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Boolean(Field):
    '''Field for storing boolean values'''
    yesNo = {'true': 'yes', 'false': 'no', True: 'yes', False: 'no'}
    trueFalse = {True: 'true', False: 'false'}

    class Layouts(Layouts):
        '''Boolean-specific layouts'''
        # Default layout (render = "checkbox") ("b" stands for "base"), followed
        # by the "grid" variant (if the field is in a group with "grid" style).
        b   = Layouts(edit=Layout('f;lrv;-', width=None), view='lf')
        g   = Layouts(edit=Layout('frvl', width=None), view='fl')
        d   = Layouts(edit=Layout('flrv;=d', width=None), view='lf')
        # *d*escription also visible on "view"
        dv  = Layouts(edit=d['edit'], view='lf-d')
        h   = Layouts(edit=Layout('flhv', width=None), view='lf')
        gd  = Layouts(edit=Layout('f;dv-', width=None), view='fl')
        # The "long" version of the previous layout (if the description is
        # long), with vertical alignment on top instead of middle.
        gdl = Layouts(edit=Layout('f;dv=', width=None), view='fl')
        # Centered layout, no description
        c   = Layouts(edit='flrv|', view='lf|')
        # Layout for radio buttons (render = "radios")
        r   = Layouts(edit='f', view='f')
        rl  = Layouts(edit='l-f', view='lf')
        rld = Layouts(edit='l-d-f', view='lf')
        grl = Layouts(edit='fl', view='fl')
        gdr = Layouts(edit=Layout('d-fv=', width=None), view='fl')

        @classmethod
        def getDefault(class_, field):
            '''Default layouts for this Boolean p_field'''
            if field.render == 'radios': return class_.r
            return class_.g if field.inGrid() else class_.b

    # The name of the index class storing values of this field in the catalog
    indexType = 'BooleanIndex'

    view = cell = Px('''
    <x>::field.getInlineEditableValue(o, value, layout, name=name)</x>
    <input type="hidden" if="masterCss"
           class=":masterCss" value=":rawValue" name=":name" id=":name"/>''')

    edit = Px('''<x var="isTrue=field.isTrue(o, name, rawValue);
                         visibleName=name + '_visible'">
     <x if="field.render == 'checkbox'">
      <input type="checkbox" name=":visibleName" id=":name"
             class=":masterCss" checked=":isTrue"
             onclick=":field.getOnChange(o, layout)"/>
     </x>
     <x if="field.render == 'radios'"
        var2="falseId='%s_false' % name;
              trueId='%s_true' % name">
      <input type="radio" name=":visibleName" id=":falseId" class=":masterCss"
             value="False" checked=":not isTrue"
             onclick=":field.getOnChange(o, layout)"/>
      <label lfor=":falseId">:_(field.labelId + '_false')</label><br/>
      <input type="radio" name=":visibleName" id=":trueId" class=":masterCss"
             value="True" checked=":isTrue"
             onclick=":field.getOnChange(o, layout)"/>
      <label lfor=":trueId">:_(field.labelId + '_true')</label>
     </x>
     <input type="hidden" name=":name" id=":'%s_hidden' % name"
            value=":isTrue and 'True' or 'False'"/>

     <script if="hostLayout" var2="x=o.Lock.set(o, user, field=field)">:\
      'prepareForAjaxSave(%s,%s,%s,%s)' % \
       (q(name), q(o.iid), q(o.url), q(hostLayout))</script></x>''',

     js='''
      updateHiddenBool = function(elem) {
        // Determine the value of the boolean field
        var value = elem.checked,
            hiddenName = elem.name.replace('_visible', '_hidden');
        if ((elem.type == 'radio') && elem.id.endsWith('_false')) value= !value;
        value = (value)? 'True': 'False';
        // Set this value in the hidden field
        document.getElementById(hiddenName).value = value;
      }''')

    search = Px('''
      <x var="valueId='%s_yes' % name">
       <input type="radio" value="True" name=":widgetName" id=":valueId"/>
       <label lfor=":valueId">:_(field.getValueLabel(True))</label>
      </x>
      <x var="valueId='%s_no' % name">
       <input type="radio" value="False" name=":widgetName" id=":valueId"/>
       <label lfor=":valueId">:_(field.getValueLabel(False))</label>
      </x>
      <x var="valueId='%s_whatever' % name">
       <input type="radio" value="" name=":widgetName" id=":valueId"
              checked="checked"/>
       <label lfor=":valueId">:_('whatever')</label>
      </x><br/>''')

    def __init__(self, validator=None, multiplicity=(0,1), default=None,
      defaultOnEdit=None, show=True, page='main', group=None, layouts=None,
      move=0, indexed=False, mustIndex=True, indexValue=None, searchable=False,
      filterField=None, readPermission='read', writePermission='write',
      width=None, height=None, maxChars=None, colspan=1, master=None,
      masterValue=None, focus=False, historized=False, mapping=None,
      generateLabel=None, label=None, sdefault=False, scolspan=1, swidth=None,
      sheight=None, persist=True, render='checkbox', inlineEdit=False,
      view=None, cell=None, edit=None, xml=None, translations=None):
        # By default, a boolean is edited via a checkbox. It can also be edited
        # via 2 radio buttons (p_render="radios").
        self.render = render
        Field.__init__(self, validator, multiplicity, default, defaultOnEdit,
          show, page, group, layouts, move, indexed, mustIndex, indexValue,
          None, searchable, filterField, readPermission, writePermission, width,
          height, None, colspan, master, masterValue, focus, historized,
          mapping, generateLabel, label, sdefault, scolspan, swidth, sheight,
          persist, inlineEdit, view, cell, edit, xml, translations)
        self.pythonType = bool

    def getValue(self, o, name=None, layout=None, single=None):
        '''Never returns "None". Returns always "True" or "False", even if
           "None" is stored in the DB.'''
        value = Field.getValue(self, o, name, layout, single)
        return False if value is None else value

    def getValueLabel(self, value):
        '''Returns the label for p_value (True or False): if self.render is
           "checkbox", the label is simply the translated version of "yes" or
           "no"; if self.render is "radios", there are specific labels.'''
        if self.render == 'radios':
            return '%s_%s' % (self.labelId, self.trueFalse[value])
        return self.yesNo[value]

    def getFormattedValue(self, o, value, layout='view', showChanges=False,
                          language=None):
        return o.translate(self.getValueLabel(value), language=language)

    def getMasterTag(self, layout):
        '''The tag driving slaves is the hidden field'''
        return (layout == 'edit') and ('%s_hidden' % self.name) or self.name

    def getOnChange(self, o, layout, className=None):
        '''Updates the hidden field storing the actual UI value for the field'''
        r = 'updateHiddenBool(this)'
        # Call the base behaviour
        base = Field.getOnChange(self, o, layout, className=className)
        return '%s; %s' % (r, base) if base else r

    def getStorableValue(self, o, value, complete=False):
        if not self.isEmptyValue(o, value):
            r = eval(value)
            return r

    def getSearchValue(self, req, value=None):
        '''Converts the raw search value from p_form into a boolean value'''
        return eval(Field.getSearchValue(self, req, value=value))

    def isSortable(self, inRefs=False):
        '''Can this field be sortable ?'''
        return True if inRefs else Field.isSortable(self) # Sortable in Refs

    def isTrue(self, o, name, dbValue):
        '''When rendering this field as a checkbox, must it be checked or
           not?'''
        req = o.req
        # Get the value we must compare (from request or from database)
        return req[name] in ('True', 1, '1') if name in req else dbValue
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
